# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from collections import namedtuple
from pathlib import Path
from typing import Callable, Optional, Type, Union

from ..data import Dataset
from pandas.io.parsers import read_csv
from .base import PackagedDataset
import shutil
import pandas as pd
import numpy as np
import h5py
import datetime


class TCGA(PackagedDataset):
    """TCGA - The Cancer Genome Atlas, packaged dataset

    The TCGA dataset captures a snapshot of The Cancer Genome Atlas, from
    the TCGA website:

    https://www.cancer.gov/about-nci/organization/ccg/research/structural-genomics/tcga
    
        "The Cancer Genome Atlas (TCGA), a landmark cancer genomics program,
        molecularly characterized over 20,000 primary cancer and matched
        normal samples spanning 33 cancer types. This joint effort between
        NCI and the National Human Genome Research Institute began in 2006,
        bringing together researchers from diverse disciplines and multiple
        institutions."

    Processing of the dataset occurs in a manner the same as that described
    by Lee in:

        Lee, Changhee, and Mihaela Schaar. "A Variational Information
        Bottleneck Approach to Multi-Omics Data Integration." International
        Conference on Artificial Intelligence and Statistics. PMLR, 2021.
    
    Although we have taken the decision to process the dataset into 10 PCA
    dimensions using linear PCA, although the number of PCA dimensions can
    be changed using the num_pca_dims argument to this constructor.
    
    Additionally no PCA can be applied and custom transformation in the form of a callable used by passing it to custom_transformation_func_and_name_tuple.

    Datasets representing clinical_decision, RPPA, miRNA, methylation,
    and mRNA are generated. These datasets are processed and saved in the
    HDF5 file format, writing files of the format
    "tcga_pca{num_pca_dims}_phenonaut.h5" and
    "tcga_pca{num_pca_dims}_metadata_phenonaut.h5" where {num_pca_dims} is
    the requested number of PCA dimensions.

    Steps undertaken in dataset preparation are as follows:
    
    #. Download the dataset. This downloads 186 .tar.gz files totalling ~20 GB. These are then extracted, taking ~32 GB.
    #. Files representing different tumour types for each view are merged, and empty columns removed. These files are then saved as intermediates.
    
    On modest 2020 hardware, processing:

    * clinical_decisions takes ~2 secs, generating 2.5 MB intermediate
    * RPPA takes ~4 secs, generating 29 MB intermediate
    * miRNA takes ~ 5 secs, generating 102 B intermediate
    * methylation takes ~12.5 mins, generating 3.6GB intermediate
    * mRNA takes 10 mins, generating 4 MB intermediate files

    Parameters
    ----------
    root : Union[Path, str]
        Local directory containing the prepared dataset. If the dataset is
        not found here and the argument download=True is not given, then an
        error is raised. If download=True and the processed dataset is
        absent, then it is downloaded the directory pointed at by the
        'raw_data' argument detailed below. If raw_data_dir is a
        non-absolute path, such as a single directory, then it is createdas
        a subdirectory of this root directory.
    download : bool, optional
        If true and the processed dataset is not found in the root
        directory, then the dataset is downloaded and processed.
        By default False.
    raw_data_dir : Optional[Union[Path, str]], optional
        If downloading and preparing the dataset, then a directory for the
        raw data may be specified. If a non-absolute location is given,
        then it is created in a subdirectory of the root directory specified
        as the first argument. Absolute paths may be used to place raw
        datafiles and intermediates in another location, such as scratch
        disks etc, by default Path("raw_data").
    rm_downloaded_data : bool
        If creating the dataset, and this is True, then downloaded data raw
        TCGA data (archives) will be deleted, by default True.
    rm_intermediates : bool
        If creating the dataset, and this is True, then intermediate data
        generated from the extraction of TCGA archives will be deleted,
        by default True.
    prediction_target: str, optional
        Often we are want to make predictions on datasets using labels or
        targets captured by TCGA and placed in the clinical_decisions
        dataframe, an example is the commonly used days_to_death column.
        This argument may be any column within that DataFrame (queryable by
        calling get_clinical_decisions_columns)
        such as days_to_death, tumor_tissue_site, gender etc additionally
        to these column names, the string 'years_to_death' can also be used,
        which will operate on days_to_death divided by 365.25,
        by default None.
    num_pca_dims: int, optional
        The number of linear principal components to use in dimensionality
        reduction. By default 10.
    vif_filter_cutoff : float, optional
        Apply a VIF (variance inflation factor) cutoff, removing features
        with a VIF score greater than this value. This has the effect or
        removing features which have a high degree of colinearity. If
        vif_filter_cutoff is None, then no vif filter is applied. A good
        default choice for this value is 5.0. If a
        custom_transformation_func_and_name_tuple as defined below is given,
        then the vif filter is ignored, to include it, you may combine ZIF
        filtering in the custom function.
    custom_transformation_func_and_name_tuple: Optional[tuple[Callable, str]]
        If PCA is not the preferred transformation to be applied to the data,
        then the user may provide their own in a tuple. The first tuple
        element should be the callable function, and the second a unique
        name/identifier which will be used to uniquely identify the saved
        dataset. Whereas datasets using the default PCA dimensionality
        reduction technique are named:
        "tcga_pca{num_pca_dims}_phenonaut.h5
        Datasets named by custom callables will be named:
        "tcga_{custom_callable_id}_phenonaut.h5
        where custom_callable_id is the second element of the
        custom_transformation_func_and_name_tuple tuple. If None, then no
        customtransformation or dimensionality reduction is performed,
        instead using the standard scalar, followed by PCA approach as
        described above. If a custom_transformation_func_and_name_tuple,
        then vif_filter_cutoff has no effect, and it is as if it is set to
        None. By default None.
    """

    # NamedTuple which helps processing of downloaded files.
    TCGA_MetadataTuple = namedtuple(
        "TCGA_MetadataTuple",
        [
            "files",  # Files containing data to be merged
            "load_csv_kwargs",  # Custom Pandas kwargs for some datasets as a row needs to be skipped.
            "output_file_name",  # Merged dataset target filename
            "header_offset",  # Header row may not be 0 in some cases.
            "treatment_id",  # All are something like "Hybridization REF", but small changes are present in raw TCGA data
            "treatment_lambda",  # TreatmentID transformation, usually convert to lowercase, sometimes remove characters.
        ],
    )

    def __init__(
        self,
        root: Union[Path, str],
        download: bool = False,
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        rm_downloaded_data: bool = True,
        rm_intermediates: bool = True,
        prediction_target: Optional[str] = None,
        num_pca_dims: Union[int, None] = 10,
        vif_filter_cutoff: Optional[float] = None,
        custom_transformation_func_and_name_tuple: Optional[tuple[Callable, str]] = None,
    ):
        # List of clinical tumor types within TCGA
        self.tumor_list = [
            "ACC",
            "BLCA",
            "BRCA",
            "CESC",
            "CHOL",
            "COAD",
            "COADREAD",
            "DLBC",
            "ESCA",
            "FPPP",
            "GBM",
            "GBMLGG",
            "HNSC",
            "KICH",
            "KIPAN",
            "KIRC",
            "KIRP",
            "LAML",
            "LGG",
            "LIHC",
            "LUAD",
            "LUSC",
            "MESO",
            "OV",
            "PAAD",
            "PCPG",
            "PRAD",
            "READ",
            "SARC",
            "SKCM",
            "STAD",
            "STES",
            "TGCT",
            "THCA",
            "THYM",
            "UCEC",
            "UCS",
            "UVM",
        ]
        super().__init__(root, raw_data_dir)
        self.prediction_target = prediction_target
        self.name = "TCGA"
        self.vif_filter_cutoff = vif_filter_cutoff
        # Handle the supply of custom transformation function

        if custom_transformation_func_and_name_tuple is not None and not isinstance(
            custom_transformation_func_and_name_tuple, tuple
        ):
            raise ValueError(
                "custom_transformation_func_and_name_tuple must be a tuple of the form tuple[Callable, str] where str is a short descriptive string to be used in the filename - such as 'UMAP' or 'LDA'"
            )

        if custom_transformation_func_and_name_tuple is None:
            self.custom_transformation_func = None
            self.processed_h5_file = (
                self.root / f"tcga_pca{num_pca_dims}_VIF{self.vif_filter_cutoff}_phenonaut.h5"
            )
            self.metadata_h5_file = (
                self.root
                / f"tcga_pca{num_pca_dims}_VIF{self.vif_filter_cutoff}_metadata_phenonaut.h5"
            )
            self.num_pca_dims = num_pca_dims
        else:
            self.custom_transformation_func = custom_transformation_func_and_name_tuple[0]
            self.processed_h5_file = (
                self.root / f"tcga_{custom_transformation_func_and_name_tuple[1]}_phenonaut.h5"
            )
            self.metadata_h5_file = (
                self.root
                / f"tcga_{custom_transformation_func_and_name_tuple[1]}_metadata_phenonaut.h5"
            )

        # If the dataset is missing, get it
        self._call_if_file_missing(
            self.processed_h5_file,
            self._make_dataset,
            {
                "download": download,
                "rm_downloaded_data": rm_downloaded_data,
                "rm_intermediates": rm_intermediates,
            },
        )

        if not self.metadata_h5_file.exists():
            raise FileNotFoundError(
                f"{self.metadata_h5_file} is not present. Something has gone wrong. Suggest deleting {self.metadata_h5_file.parent} directory and starting again"
            )

        store = pd.HDFStore(self.processed_h5_file)
        for key in store.keys():
            if key != "/clinical_decisions":
                self.register_ds_key(key[1:])
        self.register_df_key("clinical_decisions")
        self.clinical_decisions_df = store["clinical_decisions"].set_index("Hybridization REF")
        store.close()

    def _make_dataset(
        self,
        download: bool = False,
        rm_downloaded_data: bool = True,
        rm_intermediates: bool = True,
    ):
        if not download:
            print(
                "Attempting to preprocess the dataset, however, download = False. Dataset creation will fail if the files are not already present."
            )
        self.tmp_dir = self.raw_data_dir / Path("tmp")
        if not self.tmp_dir.exists():
            self.tmp_dir.mkdir(parents=True)
        if download:
            self.__download_TCGA()
        self.__decompress_TCGA()
        self.__merge_data()
        self.__perform_pca_write_h5()

        # XXX
        # If requested, remove the raw data, and processing intermediates
        # if rm_intermediates and (self.tmp_dir).exists():
        #     shutil.rmtree(self.tmp_dir)
        # if rm_downloaded_data and self.raw_data_dir.exists():
        #     shutil.rmtree(self.raw_data_dir)

    def __download_TCGA(self):
        remote_base_url = "http://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/"
        filename_suffixes = [
            "Methylation_Preprocess.Level_3.2016012800.0.0.tar.gz",
            "miRseq_Preprocess.Level_3.2016012800.0.0.tar.gz",
            "mRNAseq_Preprocess.Level_3.2016012800.0.0.tar.gz",
            "RPPA_AnnotateWithGene.Level_3.2016012800.0.0.tar.gz",
            "Clinical_Pick_Tier1.Level_4.2016012800.0.0.tar.gz",
        ]

        download_tasks = []
        print(f"{download_tasks=}")
        for tumor in self.tumor_list:
            filename_prefix = f"gdac.broadinstitute.org_{tumor}"
            for filename in [f"{filename_prefix}.{fns}" for fns in filename_suffixes]:
                if not (self.raw_data_dir / Path(filename)).exists():
                    print(f"Download task: {remote_base_url}{tumor}/20160128/{filename}")
                    download_tasks.append(
                        (
                            f"{remote_base_url}{tumor}/20160128/{filename}",
                            self.raw_data_dir / filename,
                        )
                    )
        print(f"Downloading {len(download_tasks)} files")
        # According to the format and patterns of deposited TCGA data, the 4 files
        # below should exist, but they are not present on the server.  Including
        # them as non-critical allows checking if they are available without raising
        # an exception when they are not downloaded.
        non_critical_urls = [
            "http://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/FPPP/20160128/gdac.broadinstitute.org_FPPP.Methylation_Preprocess.Level_3.2016012800.0.0.tar.gz",
            "http://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/FPPP/20160128/gdac.broadinstitute.org_FPPP.mRNAseq_Preprocess.Level_3.2016012800.0.0.tar.gz",
            "http://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/FPPP/20160128/gdac.broadinstitute.org_FPPP.RPPA_AnnotateWithGene.Level_3.2016012800.0.0.tar.gz",
            "http://gdac.broadinstitute.org/runs/stddata__2016_01_28/data/LAML/20160128/gdac.broadinstitute.org_LAML.RPPA_AnnotateWithGene.Level_3.2016012800.0.0.tar.gz",
        ]
        self._batch_download(download_tasks, non_critical_urls=non_critical_urls)

    def __decompress_TCGA(self):
        # Extract compressed files
        for file in [
            f
            for f in self.raw_data_dir.glob("*.gz")
            if not Path(str(f)[: -len(".tar.gz")]).exists()
        ]:
            # Some files like gdac.broadinstitute.org_LIHC.RPPA_AnnotateWithGene.Level_3.2016012800.0.0.tar.gz
            # do not unpack themselves to the expected
            # Expected: gdac.broadinstitute.org_LIHC.RPPA_AnnotateWithGene.Level_3.2016012800.0.0
            # Actual  : gdac.broadinstitute.org_LIHC.RPPA_AnnotateWithGene.Level_3.2016071400.0.0
            # So a single attempt to extract all is good enough.
            print(f"Extracing {file}")
            shutil.unpack_archive(file, file.parent)

    def __merge_data(self):
        tcga_raw_data_metadata = {
            "clinical_decision": self.TCGA_MetadataTuple(
                list(self.raw_data_dir.glob("**/*.clin.merged.picked.txt")),
                {},
                "clinical_decisions.csv",
                1,
                "Hybridization REF",
                lambda x: x.lower(),
            ),
            "RPPA": self.TCGA_MetadataTuple(
                list(self.raw_data_dir.glob("**/*.rppa.txt")),
                {},
                "RPPA.csv",
                1,
                "Composite.Element.REF",
                lambda x: x.lower()[:12],
            ),
            "miRNA": self.TCGA_MetadataTuple(
                list(self.raw_data_dir.glob("**/*.miRseq_RPKM_log2.txt")),
                {},
                "miRNAseq_RPKM_log2.csv",
                1,
                "gene",
                lambda x: x.lower()[:-3],
            ),
            "methylation": self.TCGA_MetadataTuple(
                list(self.raw_data_dir.glob("**/*.meth.by_mean.data.txt")),
                {"skiprows": [1]},
                "methylation.csv",
                2,
                "Hybridization REF",
                lambda x: x.lower()[:-3],
            ),
            "mRNA": self.TCGA_MetadataTuple(
                list(self.raw_data_dir.glob("**/*.uncv2.mRNAseq_RSEM_normalized_log2.txt")),
                {},
                "mRNAseq_RSEM.csv",
                1,
                "gene",
                lambda x: x.lower()[:-3],
            ),
        }

        # Make single CSV files
        for dataset_name, metadata in tcga_raw_data_metadata.items():
            print(
                f"{datetime.datetime.now().strftime('%H:%M:%S')}: Processing {dataset_name} ... ",
                end="",
            )
            self.__merge_TCGA_files(metadata)
            print(f"{datetime.datetime.now().strftime('%H:%M:%S')}")

    def __merge_TCGA_files(self, dataset_metadata: TCGA_MetadataTuple):

        if not (self.tmp_dir / dataset_metadata.output_file_name).exists():
            TARGET_TREATMENT_ID = "Hybridization REF"
            dfs = []
            for file in dataset_metadata.files:
                tumor = str(file.name)[: file.name.find(".")]

                print(f"{tumor}")

                tmp = pd.read_csv(file, sep="\t", header=[0], **dataset_metadata.load_csv_kwargs)
                tmp = tmp.T.reset_index()
                tmp.columns = tmp.iloc[0, 0:]
                tmp = tmp.iloc[dataset_metadata.header_offset :, :].reset_index(drop=True)
                tmp[TARGET_TREATMENT_ID] = tmp[dataset_metadata.treatment_id].apply(
                    dataset_metadata.treatment_lambda
                )
                tmp["disease_code"] = tumor
                dfs.append(tmp)

            merged_df = (
                pd.concat(dfs)
                .drop(columns=["Composite Element REF"], errors="ignore")
                .replace(r"^\s*$", np.nan, regex=True)
            )
            if dataset_metadata.treatment_id != TARGET_TREATMENT_ID:
                merged_df = merged_df.drop(columns=[dataset_metadata.treatment_id], errors="ignore")
            del dfs
            merged_df.to_csv(self.tmp_dir / dataset_metadata.output_file_name, index=False)
            del merged_df

    def __perform_pca_write_h5(self):
        from sklearn.decomposition import PCA
        from sklearn.preprocessing import StandardScaler

        store = pd.HDFStore(self.processed_h5_file, "w")
        metadata_hdf = h5py.File(self.metadata_h5_file, "w")
        print("Running dimensionality reduction")

        for file_path in [
            self.tmp_dir / f
            for f in [
                "RPPA.csv",
                "miRNAseq_RPKM_log2.csv",
                "mRNAseq_RSEM.csv",
                "methylation.csv",
            ]
        ]:
            if self.custom_transformation_func is not None:
                self.custom_transformation_func(file_path, store)
            else:
                dataset_type = (
                    str(file_path.name).split("_")[0].replace(".csv", "").replace("seq", "")
                )
                index_column = "Hybridization REF"
                print(f"Performing dimensionality reduction on {file_path}")
                df = pd.read_csv(file_path)
                df = df[df[index_column].astype(str).str.startswith("tcga")]
                df = df.drop_duplicates(subset=[index_column])
                df = df[np.asarray(list(df))[df.isna().sum(axis=0) == 0]]
                df = df.set_index(index_column)

                features = sorted(
                    [f for f in df.columns.values if f not in [index_column, "disease_code"]]
                )

                if self.vif_filter_cutoff is not None:
                    from phenonaut.transforms.preparative import RemoveHighestCorrelatedThenVIF

                    tmp_ds = Dataset("tmp_df", df, {"features": features})
                    print("Num features in dataset:", len(tmp_ds.features))
                    # If there are too many features, then we cannot analyse all
                    # correlations and must abandon
                    if len(tmp_ds.features) <= 1000:
                        r_high_then_vif = RemoveHighestCorrelatedThenVIF(verbose=True)
                        r_high_then_vif.filter(
                            tmp_ds, vif_cutoff=self.vif_filter_cutoff, drop_columns=True
                        )
                        features = tmp_ds.features
                    df = tmp_ds.df
                sc = StandardScaler()
                df_features = sc.fit_transform(df[features])
                h5_root_group = metadata_hdf.create_group(dataset_type)
                g_scaler = h5_root_group.create_group("standardscaler")
                g_scaler.create_dataset("mean_", data=sc.mean_)
                g_scaler.create_dataset("scale_", data=sc.scale_)

                g_pca = h5_root_group.create_group("pca")

                pca = PCA(n_components=self.num_pca_dims)
                pcaspace = pca.fit_transform(df_features)
                pca_features = [f"PC{n+1}" for n in range(self.num_pca_dims)]

                pca_df = pd.DataFrame(data=pcaspace, columns=pca_features)
                pca_df["disease_code"] = df["disease_code"].astype(str).values
                pca_df[index_column] = df.index
                pca_df = pca_df[pca_df[index_column].astype(str).str.startswith("tcga")]

                pca_df.set_index(index_column, inplace=True)
                h5path = f"{dataset_type}"
                store[h5path] = pca_df

                this_pca_grp = g_pca.create_group(str(self.num_pca_dims))
                this_pca_grp.create_dataset("n_components", data=self.num_pca_dims)
                this_pca_grp.create_dataset("components", data=pca.components_)
                this_pca_grp.create_dataset("mean", data=pca.mean_)
        store["clinical_decisions"] = pd.read_csv(
            self.tmp_dir / "clinical_decisions.csv", low_memory=False
        )
        store.close()
        metadata_hdf.close()

    def get_df(self, key: str) -> pd.DataFrame:
        """Get supporting DataFrame

        Parameters
        ----------
        key : str
            Key of pd.DataFrame.

        Returns
        -------
        pd.DataFrame
            Requested pd.DataFrame from h5 store
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        store.close()
        return df

    def get_ds(self, key: str) -> Dataset:
        """Get Dataset

        Parameters
        ----------
        key : str
            Key of Phenonaut Dataset.

        Returns
        -------
        Dataset
            Requested Phenonaut Dataset from h5 store, with the correctly set
            features and metadata
        """
        store = pd.HDFStore(self.processed_h5_file)
        if key.startswith("/"):
            key = key[1:]
        df = store["/" + key]
        features = [f"PC{d}" for d in range(1, self.num_pca_dims + 1)]

        if self.prediction_target is not None:
            if self.prediction_target == "years_to_death":
                df = self.add_clinical_decision_data_to_df(
                    df, "days_to_death", self.prediction_target, lambda x: x / 365.25
                )
            elif self.prediction_target == "survives_1_year":
                df = self.add_clinical_decision_data_to_df(
                    df,
                    "days_to_death",
                    self.prediction_target,
                    lambda x: 1 if (x / 365.25) >= 1 else 0,
                    dtype=int,
                )
            else:
                df = self.add_clinical_decision_data_to_df(df, self.prediction_target)

        ds = Dataset(key, df, {"features": features})
        if self.prediction_target is not None:
            ds._metadata["prediction_target"] = self.prediction_target
        store.close()
        return ds

    def add_clinical_decision_data_to_df(
        self,
        df: pd.DataFrame,
        clinical_decision_column: str,
        new_df_column_name: Optional[str] = None,
        custom_func: Optional[Callable[[pd.Series], pd.Series]] = None,
        remove_incomplete_rows: bool = True,
        dtype: Type = float,
    ) -> pd.DataFrame:
        """Merge a field from TCGA clinical_decisions.

        The TCGA dataset comes with a clinical_decisions DataFrame specifying
        things like patient age, days to death etc. As years to death can be a
        prediction target, we need a way to add this to our multiomics
        DataSets/DataFrames. This function merges the clinical_decisions
        information based on the "Hybridization REF" index.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame to which the information will be added
        clinical_decision_column : str
            Column in clinical_decitions that will be added
        new_df_column_name : Optional[str], optional
            The new column may have a new name. If None, then the clinical_decision_column
            is used, by default None.
        custom_func : Optional[Callable[[pd.Series], pd.Series]], optional
            Optionally apply a transformation to the newly added data. It can
            be useful to pass a lambda here, which can enable a simple way to
            convert days to years, such as : 'lambda x: x/365.'. If None, then
            no transformation is applied to the new column, by default None.
        remove_incomplete_rows : bool, optional
            If True, rows containing missing data in the new column (by virtue
            of a pd.na being present) are removed, by default True.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the new column.
        """
        if new_df_column_name is None:
            new_df_column_name = clinical_decision_column

        cd_series = self.clinical_decisions_df[clinical_decision_column]
        if custom_func is not None:
            cd_series = cd_series.map(custom_func)
        df = df.join(cd_series).rename(columns={clinical_decision_column: new_df_column_name})

        if remove_incomplete_rows:
            df = df[~df[new_df_column_name].isin([np.nan, np.inf, -np.inf])]
        if dtype is not None:
            df[new_df_column_name] = df[new_df_column_name].astype(dtype)
        return df



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Optional, Union
import yaml
from pathlib import Path
import json
from phenonaut.utils import load_dict

class PhenonautVisualisation:
    """Base class for Phenonaut visualisations.
    
    Constructor allows supply of a plot_config dictionary, which is then stored
    in object under the name 'config'.  Out of this dictionary, if it exists,
    then a nested 'plot_markers' dictionary is extracted and placed into the
    class' markers member variable.


    Parameters
    ----------
    config : Optional[Union[Path, str, dict]], optional
        Configuration dictionary, which will be stored in config member
        variable. If None, then an empty dictionary is initialised as the
        config member variable. By default, None
    """    
    
    def __init__(self, plot_config:Optional[Union[Path, str, dict]]=None):
        markers = {}
        self.config=load_dict(plot_config, cast_none_to_dict=True)
        for k, v in self.config.get("plot_markers", {}).items():
            self.markers[k] = v
            



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from phenonaut import data
from numpy import isin
from pandas.errors import DataError

from phenonaut.data import Dataset
from typing import Callable, Optional, Union
from sklearn.decomposition import PCA as Sklearn_PCA
from sklearn.manifold import TSNE as Sklearn_TSNE
import umap
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as sklearn_LDA
import numpy as np
from matplotlib import pyplot as plt
from pathlib import Path
import pandas as pd

from phenonaut.phenonaut import Phenonaut
from phenonaut.transforms.transformer import Transformer


class PCA(Transformer):
    """Principal Component Analysis (PCA) dimensionality reduction.

    Can be instantiated and called like:

    .. code-block:: python

        pca=PCA()
        pca(dataset)

    Parameters
    ----------
    dataset : Dataset
        The Phenonaut dataset on which to apply PCA
    new_feature_names : Union[list[str], str]
        List of strings containing the names for the new features. Can also
        be just a single string, which then has numerical suffixes attached
        enumerating the number of new features generated. By default "PCA"
    ndims : int, optional
        Number of dimensions to embed the data into, by default 2
    """

    def __init__(self, new_feature_names="PC", ndims:int=2):
        super().__init__(
            Sklearn_PCA,
            new_feature_names=new_feature_names,
            transformer_name="SciKitPCA",
            constructor_kwargs={'n_components':ndims}
        )
        self.ndims = ndims

    def __call__(
        self,
        dataset: Union[Dataset, Phenonaut],
        ndims: int = 2,
        new_feature_names:Union[list[str], str]="PC",
        group_by: Optional[Union[str, list[str]]] = None,
        center_on_perturbation_id: Optional[str] = None,
        centering_function:Callable=np.median,

        fit_perturbation_ids: Optional[Union[str, list]] = None,
        fit_query:Optional[str]=None,
        explain_variance_in_features: bool = False,
    ):
        """Principal Component Analysis (PCA) dimensionality reduction

        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object containing one dataset
            on which to apply PCA.
        ndims : int
            Number of dimensions to embed the data into, by default 2
        new_feature_names : Union[list[str], str]
            List of strings containing the names for the new features. Can also
            be just a single string, which then has numerical suffixes attached
            enumerating the number of new features generated. By default "PCA"
        group_by : Optional[Union[str, list[str]]], optional
            Often we would like to apply transformations on a plate-by-plate
            basis. This group_by argument allows definition of a string which
            is used to identify columns with which unique values define a group
            or plate. It works with the pandas group_by function and accepts
            a list of strings if multiple columns must be used to define groups.
            By default None.
        center_on_perturbation_id : Optional[str], optional
            Optionally, recentre the PCA on a named perturbation. Should have
            pertubation_column set within the dataset for this option, by
            default None.
        centering_function : Callable
            Used with center_on_perturbation, this function is applied to
            all data for matching perturbations. By default, we use the median
            of perturbations. This behavior can be overridden by supplying a
            different function here. By default np.median.
        fit_perturbation_ids : Optional[Union[str, list]], optional
            If only a subset of the data should be used for fitting, then
            pertubations for fitting may be listed here. If none, then
            every datapoint is used for fitting, by default None.
        fit_query : str, optional
            A pandas style query may be supplied to perform fitting. By default
            None.
        explain_variance_in_features : bool, optional
            If True, then the percentage explained variance of each PCA
            dimension is included in the new PCA descriptor feature name.
            Overrides new_feature_names. By default False.
        """
        if self.ndims!=ndims:
            super().__init__(
                Sklearn_PCA,
                new_feature_names=new_feature_names,
                transformer_name="SciKitPCA",
                constructor_kwargs={'n_components':ndims},
            )
            self.ndims = ndims
        self.fit_transform(
            dataset,
            group_by=group_by,
            fit_perturbation_ids=fit_perturbation_ids,
            fit_query=fit_query,
            new_feature_names=new_feature_names,
            method_name="SciKitPCA",
            center_on_perturbation_id=center_on_perturbation_id,
            centering_function=centering_function,
        )
        if explain_variance_in_features:
            if isinstance(self._method, list):
                print("Warning, adding variance explained by features, however, group_by was used. Explaining variance using the first fitted PCA")
                pca_variance_ratio=self._method[0].explained_variance_ratio_
            else:
                pca_variance_ratio=self._method.explained_variance_ratio_

            explained_var_features=[
                    f"PC{x+1} ({expvariance*100:.2f} % explained variance)"
                    for x, expvariance in zip(
                        range(self.ndims), pca_variance_ratio
                    )
                ]

            column_rename_dict = {k:v for k,v in zip(dataset.features,explained_var_features)}
            dataset.df.rename(columns=column_rename_dict, inplace=True)
            dataset.features=explained_var_features, "Added variance to features"
            
    def save_scree_plot(
        self, output_filename: Optional[Union[str, Path]] = None, title="Scree plot"
    ):
        """Produce a Scree plot showing ndims vs explained variance

        Parameters
        ----------
        output_filename : Optional[Union[str, Path]]
            Output filename (ending in .png) indicating where the Scree plot should
            be saved. If None, then the plot is displayed interactively. By default, None.
        title : str, optional
            Plot title, by default "Scree plot"

        Raises
        ------
        DataError
            PCA must have been fitted before a Scree plot can be made.
        """
        if self._method is None:
            raise DataError("Cant make scree plot, nothing fitted")
        if isinstance(self._method, list):
            pca_object=self._method[-1]
        else:
            pca_object=self._method
        fig, ax = plt.subplots(1)
        ax.plot(
            range(1, pca_object.explained_variance_ratio_.shape[0] + 1),
            np.cumsum(pca_object.explained_variance_ratio_),
            label="Cumulative explained variance",
            marker="x",
        )
        ax.plot(
            range(1, pca_object.explained_variance_ratio_.shape[0] + 1),
            pca_object.explained_variance_ratio_,
            label="Explained variance",
            marker="x",
        )

        ax.set_xlabel("Feature")
        ax.set_ylabel("Fraction explained variance")
        ax.set_xlim((0.25, self.pca.explained_variance_ratio_.shape[0]))
        ax.set_ylim((0, 1))
        ax.set_title(title)
        plt.xticks(range(1, self.pca.explained_variance_ratio_.shape[0] + 1))
        plt.tight_layout()
        plt.grid()
        ax.legend()
        if output_filename is None:
            plt.show()
        else:
            if isinstance(output_filename, str):
                output_image_path = Path(output_filename)
            plt.savefig(output_filename)

        

class TSNE(Transformer):
    """t-SNE dimensionality reduction

    Can be instantiated and called like:

    .. code-block:: python

        tsne=TSNE()
        tsne(dataset)

    Parameters
    ----------
    dataset : Dataset
        The Phenonaut dataset on which to apply the transformation
    new_feature_names : Union[list[str], str]
        List of strings containing the names for the new features. Can also
        be just a single string, which then has numerical suffixes attached
        enumerating the number of new features generated. By default "TSNE"
    ndims : int, optional
        Number of dimensions to embed the data into, by default 2
    """

    def __init__(self, new_feature_names="TSNE", ndims:int=2):
        
        constructor_kwargs={'n_components':ndims}
        
        if ndims > 3:
            constructor_kwargs['method'] = "exact"
        else:
            constructor_kwargs['method'] = "barnes_hut"
        
        super().__init__(
            Sklearn_TSNE,
            new_feature_names=new_feature_names,
            transformer_name="t-SNE",
            constructor_kwargs=constructor_kwargs
        )
        self.ndims = ndims

    def __call__(
        self,
        dataset: Union[Dataset, Phenonaut],
        ndims: int = 2,
        new_feature_names:Union[list[str], str]="tSNE",
        group_by: Optional[Union[str, list[str]]] = None,
        center_on_perturbation_id: Optional[str] = None,
        centering_function:Callable=np.median,
    ):
        """t-SNE dimensionality reduction

        Once instantiated, can be called directly, like:

        .. code-block:: python

            tsne=TSNE()
            tsne(dataset)

        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object containing one dataset
            on which to apply t-SNE.
        ndims : int
            Number of dimensions to embed the data into, by default 2
        new_feature_names : Union[list[str], str]
            List of strings containing the names for the new features. Can also
            be just a single string, which then has numerical suffixes attached
            enumerating the number of new features generated. By default "tSNE"
        group_by : Optional[Union[str, list[str]]], optional
            Often we would like to apply transformations on a plate-by-plate
            basis. This group_by argument allows definition of a string which
            is used to identify columns with which unique values define a group
            or plate. It works with the pandas group_by function and accepts
            a list of strings if multiple columns must be used to define groups.
            By default None.
        center_on_perturbation_id : Optional[str], optional
            Optionally, recentre the t-SNE on a named perturbation. Should have
            pertubation_column set within the dataset for this option, by
            default None.
        centering_function : Callable
            Used with center_on_perturbation, this function is applied to
            all data for matching perturbations. By default, we use the median
            of perturbations. This behavior can be overridden by supplying a
            different function here. By default np.median.
        """
        if self.ndims!=ndims:
            constructor_kwargs={'n_components':ndims}
            if ndims > 3:
                constructor_kwargs['method'] = "exact"
            else:
                constructor_kwargs['method'] = "barnes_hut"
            super().__init__(
                TSNE,
                new_feature_names=new_feature_names,
                transformer_name="t-SNE",
                constructor_kwargs=constructor_kwargs
            )
            self.ndims=ndims

        self.fit_transform(
            dataset,
            group_by=group_by,
            new_feature_names=new_feature_names,
            method_name="t-SNE",
            center_on_perturbation_id=center_on_perturbation_id,
            centering_function=centering_function,
        )



class UMAP(Transformer):
    """UMAP dimensionality reduction

    Can be instantiated and called like:

    .. code-block:: python

        umap=UMAP()
        umap(dataset)

    Parameters
    ----------
    dataset : Dataset
        The Phenonaut dataset on which to apply the transformation
    new_feature_names : Union[list[str], str]
        List of strings containing the names for the new features. Can also
        be just a single string, which then has numerical suffixes attached
        enumerating the number of new features generated. By default "UMAP"
    ndims : int, optional
        Number of dimensions to embed the data into, by default 2
    """

    def __init__(self, new_feature_names="UMAP", ndims:int=2):
        super().__init__(
            umap.UMAP,
            new_feature_names=new_feature_names,
            transformer_name="UMAP",
            constructor_kwargs={'n_components':ndims}
        )
        self.ndims = ndims

    def __call__(
        self,
        dataset: Union[Dataset, Phenonaut],
        ndims: int = 2,
        new_feature_names:Union[list[str], str]="UMAP",
        group_by: Optional[Union[str, list[str]]] = None,
        center_on_perturbation_id: Optional[str] = None,
        centering_function:Callable=np.median,
    ):
        """UMAP dimensionality reduction

        Once instantiated, can be called directly, like:

        .. code-block:: python

            uamp=UMAP()
            umap(dataset)

        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object containing one dataset
            on which to apply UMAP.
        ndims : int
            Number of dimensions to embed the data into, by default 2
        new_feature_names : Union[list[str], str]
            List of strings containing the names for the new features. Can also
            be just a single string, which then has numerical suffixes attached
            enumerating the number of new features generated. By default "UMAP"
        group_by : Optional[Union[str, list[str]]], optional
            Often we would like to apply transformations on a plate-by-plate
            basis. This group_by argument allows definition of a string which
            is used to identify columns with which unique values define a group
            or plate. It works with the pandas group_by function and accepts
            a list of strings if multiple columns must be used to define groups.
            By default None.
        center_on_perturbation_id : Optional[str], optional
            Optionally, recentre the UMAP on a named perturbation. Should have
            pertubation_column set within the dataset for this option, by
            default None.
        centering_function : Callable
            Used with center_on_perturbation, this function is applied to
            all data for matching perturbations. By default, we use the median
            of perturbations. This behavior can be overridden by supplying a
            different function here. By default np.median.
        """
        if self.ndims!=ndims:
            super().__init__(
                umap.UMAP,
                new_feature_names=new_feature_names,
                transformer_name="UMAP",
                constructor_kwargs={'n_components':ndims}
            )
            self.ndims=ndims

        self.fit_transform(
            dataset,
            group_by=group_by,
            new_feature_names=new_feature_names,
            method_name="UMAP",
            center_on_perturbation_id=center_on_perturbation_id,
            centering_function=centering_function,
        )


class LDA:
    """LDA dimensionality reduction

        Once instantiated, can be called like:

        .. code-block:: python

            lda=LDA()
            lda(dataset)


        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object containing one dataset
            on which to apply the transformation.
        ndims : int, optional
            Number of dimensions to embed the data into, by default 2
        center_on_perturbation_id : Optional[str], optional
            Optionally, recentre the embedding space on a named perturbation. Should have
            pertubation_column set within the dataset for this option, by
            default None.
        center_by_median : bool, optional
            By default, any dataset centering will be performed on the median
            of controls or perturbations. If this argument is False, then
            centering is performed on the mean, by default True.
        predict_proba : bool
            If True, then probabilities of each datapoint belonging to every
            other class is calculated and used in place of output features.
        """
    def __init__(self):
        self.ndims = 2
        self.lda = None

    def __call__(
        self,
        dataset: Dataset,
        ndims=2,
        center_on_perturbation_id=None,
        center_by_median: bool = True,
        predict_proba:bool=False,
    ):
        """LDA dimensionality reduction

        Once instantiated, can be called like:

        .. code-block:: python

            lda=LDA()
            lda(dataset)


        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object containing one dataset
            on which to apply the transformation.
        ndims : int, optional
            Number of dimensions to embed the data into, by default 2
        center_on_perturbation_id : Optional[str], optional
            Optionally, recentre the embedding space on a named perturbation. Should have
            pertubation_column set within the dataset for this option, by
            default None.
        center_by_median : bool, optional
            By default, any dataset centering will be performed on the median
            of controls or perturbations. If this argument is False, then
            centering is performed on the mean, by default True.
        predict_proba : bool
            If True, then probabilities of each datapoint belonging to every
            other class is calculated and used in place of output features.
        """
        if isinstance(dataset, Phenonaut):
            if len(dataset.datasets) == 1:
                dataset = dataset.ds
            else:
                raise ValueError(
                    "Phenonaut object with more than one dataset passed to fit, please pass the dataframe to apply the transform to it"
                )
        lda_feature_column_names = [f"LDA{x+1}" for x in range(ndims)]
        fit_perturbation_ids = list(dataset.df[dataset.perturbation_column].unique())

        self.lda = sklearn_LDA(n_components=ndims)
        X, y = (
            dataset.df.loc[:, dataset.features],
            dataset.df.loc[:, dataset.perturbation_column],
        )
        self.lda.fit(X, y)

        dataset.df.loc[:, lda_feature_column_names] = self.lda.transform(X)
        dataset.features = (
            lda_feature_column_names,
            f"Performed LDA, ndims={ndims}",
        )
        if center_on_perturbation_id is not None:
            if isinstance(center_on_perturbation_id, str):
                if center_by_median:
                    center_on_coords = np.median(
                        dataset.df.query(
                            f"{dataset.perturbation_column} =='{center_on_perturbation_id}'"
                        )[dataset.features],
                        axis=0,
                    )
                else:
                    center_on_coords = np.mean(
                        dataset.df.query(
                            f"{dataset.perturbation_column} =='{center_on_perturbation_id}'"
                        )[dataset.features],
                        axis=0,
                    )
            else:
                center_on_coords = center_on_perturbation_id
            dataset.df.loc[:, dataset.features] = (
                dataset.df.loc[:, dataset.features] - center_on_coords
            )
            dataset.features = (
                dataset.features,
                f"Centered on {center_on_perturbation_id}, {center_by_median=}",
            )

        if predict_proba:
            dataset.df[[f"proba_{c}" for c in self.lda.classes_]] = self.lda.predict_proba(
                dataset.df[dataset.features]
            )
            for index, row in dataset.df.iterrows():
                correct_proba = row[f"proba_{row['Edit']}"]
                dataset.df.loc[index, "Correct_proba"] = correct_proba

            dataset.features = (
                [f"proba_{c}" for c in self.lda.classes_],
                "LDA prediction probability",
            )
            dataset.df.to_csv("a.csv")
        else:
            dataset.features = (
                lda_feature_column_names,
                f"Performed LDA, ndims={ndims}",
            )

    def make_scree_plot(self, output_filename: None, title="Scree plot"):
        """Produce a Scree plot showing ndims vs explained variance

        Parameters
        ----------
        output_filename : Optional[Union[str, Path]]
            Output filename (ending in .png) indicating where the Scree plot should
            be saved. If None, then the plot is displayed interactively. By default, None.
        title : str, optional
            Plot title, by default "Scree plot"

        Raises
        ------
        DataError
            LDA must have been fitted before a Scree plot can be made.
        """
        if self.lda is None:
            raise DataError("Cant make scree plot, nothing fitted")
        fig, ax = plt.subplots(1)
        ax.plot(
            range(1, self.lda.explained_variance_ratio_.shape[0] + 1),
            np.cumsum(self.lda.explained_variance_ratio_),
            label="Cumulative explained variance",
            marker="x",
        )
        ax.plot(
            range(1, self.lda.explained_variance_ratio_.shape[0] + 1),
            self.lda.explained_variance_ratio_,
            label="Explained variance",
            marker="x",
        )

        ax.set_xlabel("Feature")
        ax.set_ylabel("Fraction explained variance")

        ax.set_xlim((0.25, self.lda.explained_variance_ratio_.shape[0]))
        ax.set_ylim((0, 1))
        ax.set_title(title)
        plt.xticks(range(1, self.lda.explained_variance_ratio_.shape[0] + 1))
        plt.tight_layout()
        plt.grid()
        ax.legend()
        if output_filename is None:
            plt.show()
        else:
            if isinstance(output_filename, str):
                output_image_path = Path(output_filename)
            plt.savefig(output_filename)



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

import warnings
from random import random
from pandas.errors import DataError
from phenonaut import data
from phenonaut.data import Dataset
from typing import Callable, Iterable, List, Optional, Union
from pathlib import Path
import pandas as pd
from itertools import combinations as itertools_combinations
from sklearn.utils import Bunch
from phenonaut.packaged_datasets.base import PackagedDataset
import pickle
import gzip
from phenonaut.utils import check_path
from hashlib import sha256
from base64 import b64encode


class Phenonaut:
    """Phenonaut object constructor

    Holds multiple datasets of different type, applys transforms, load and
    tracking operations.

    May be initialised with:

    * Phenonaut Datasets
    * Phenonaut PackageDataset
    * Scikit Bunch
    * pd.DataFrame

    by passing the object as an optional dataset argument.

    Parameters
    ----------
    dataset : Optional[Union[Dataset, list[Dataset], PackagedDataset, Bunch,
        pd.DataFrame, Path, str]], optional
        Initialise Phenonaut object with a Dataset, list of datasets, or
        PackagedDataset, by default None.
    name : str
        A name may be given to the phenonaut object. This is useful in
        naming collections of datasets. For example, The Cancer Genome Atlas
        contains 4 different views on tumors - mRNA, miRNA, methylation and
        RPPA, collectively, these 4 datasets loaded into a phenonaut object
        may be named 'TCGA' - or 'The Cancer Genome Atlas dataset'. If set
        to None, then the phenonaut object takes the name "Phenonaut data",
        however, not in the case where construction of the object occurs
        with a phenonaut packaged dataset or already named phenonaut object,
        where it takes the name of the passed object/dataset.
    kind : Optional[str]
        Instead of providing metadata, some presets are available, which make reading
        in things like DRUG-Seq easier. This argument only has an effect when reading
        in a raw data file, like CSV or H5 and directs Phenonaut to use a predefind set
        of parameters/transforms. If used as well as metadata, then the preset metadata
        dictionary from the kind argument is first loaded, then updated with anything in
        the metadata dictionary, this therefore allows overriding specific presets present
        in kind dictionaries. Available 'kind' dictionaries may be listed by examining:
        phenonaut.data.recipes.recipes.keys()
    packaged_dataset_name_filter : Optional[Union[list[str], str]], optional
        If a PackagedDataset is supplied for the data argument, then import
        only datasets from it named in the name_filter argument. If None,
        then all PackagedDataset datasets are imported.  Can be a single
        string or list of strings. If None, and PackagedDataset is supplied,
        then all Datasets are loaded. Has no effect if data is not a
        PackagedDataset, by default None.
    metadata : Optional[Union[dict, list[dict]]]
        Used when a pandas DataFrame is passed to the constructor of the
        phenonaut object. Metadata typically contains features or
        feature_prefix keys telling Phenonaut which columns should be
        treated as Dataset features. Can also be a list of metadata
        dicitonaries if a list of pandas DataFrames are supplied to the
        constructor. Has no effect if the type of dataset passed is not a
        pandas DataFrame or list of pandas DataFrames. If a list of pandas
        DataFrames is passed to data but only one metadata dictionary is
        given, then this dictionary is applied to all DataFrames. By default
        None.
    dataframe_name : Optional[Union[dict, list[dict]]]
        Used when a pandas DataFrame, or str, or Path to a CSV file is passed to
        the constructor of the phenonaut object. Optional name to give to the
        dataset object constructed from the pandas DataFrame. If multiple
        DataFrames are given in a list, then this dataframe_name argument can be
        a list of strings as names to assign to the new Dataset objects.
    init_hash : Optional[Union[str, bytes]]
        Cryptographic hashing within Phenonaut can be initialised with a
        starting/seed hash. This is useful in the creation of
        blockchain-like chains of hashes. In environments where timestamping
        is unavailable, hashes may be published and then used as input to
        subsequent experiments. Building up a provable chain along the way.
        By default None, implying an empty bytes array.
    """

    def __init__(
        self,
        dataset: Optional[
            Union[Dataset, list[Dataset], PackagedDataset, Bunch, pd.DataFrame, Path, str]
        ] = None,
        name: str = "Phenonaut object",
        kind: Optional[str] = None,
        packaged_dataset_name_filter: Optional[Union[list[str], str]] = None,
        metadata: Optional[Union[dict, list[dict]]] = None,
        dataframe_name: Optional[Union[str, list[str]]] = None,
        init_hash: Optional[Union[str, bytes]] = None,
    ):

        if init_hash is None:
            self.init_hash = b""
        else:
            if isinstance(init_hash, bytes):
                self.init_hash = init_hash
            elif isinstance(init_hash, str):
                self.init_hash = init_hash.encode("utf-8")
            else:
                raise ValueError(f"The given argument init_hash was not a bytes array or a string")
        self.datasets = []
        self.name = name
        if dataset is None:
            return
        if isinstance(dataset, Phenonaut):
            for ds in dataset.datasets:
                self.datasets.append(ds.copy())
            if name is None and dataset.name is not None:
                self.name = dataset.name
            return

        # If scikit bunch
        if isinstance(dataset, Bunch):
            # Check required keys are in Bunch
            required_bunch_keys = ["data", "feature_names", "target"]
            missing_keys = set(required_bunch_keys) - set(dataset.keys())
            if len(missing_keys) > 0:
                raise ValueError(
                    f"The supplied bunch did not contain the required keys ({required_bunch_keys}), containing only: {dataset.keys()}, missing keys were: {missing_keys}"
                )
            # Construct pd.DataFrame from the bunch.
            df = pd.concat(
                [
                    pd.DataFrame(dataset.data, columns=dataset.feature_names),
                    pd.DataFrame(dataset.target, columns=["target"]),
                ],
                axis=1,
            )
            # Append to datasets, setting features and name.
            if dataframe_name is None:
                self.datasets.append(
                    Dataset(dataset_name="sklearnBunch", input_file_path=df, metadata={"features": dataset.feature_names})
                )
            else:
                self.datasets.append(
                    Dataset(dataset_name=dataframe_name, input_file_path=df, metadata={"features": dataset.feature_names})
                )

            if name is not None:
                self.name = name
            return

        # If Path or string, then read it in
        if isinstance(dataset, (Path, str)):
            if isinstance(dataset, str):
                dataset = Path(dataset)
            if dataframe_name is None:
                dataframe_name = str(dataset)
            dataset = Dataset(dataset_name=dataframe_name, input_file_path=dataset, metadata=metadata, kind=kind)

        # If its a Dataset, or list of datasets, directly add it to self.datasets.
        if isinstance(dataset, (Dataset, pd.DataFrame)):
            dataset = [dataset]
        if isinstance(dataset, list):

            ds_types = list(set(map(type, dataset)))

            if len(ds_types) > 1:
                raise ValueError(
                    f"If passing a list for the data argument, all elements must be of the same type, they were {list(map(type, dataset))}"
                )
            if ds_types[0] is not pd.DataFrame and ds_types[0] is not Dataset:
                raise ValueError(
                    f"List containing {ds_types} passed, but lists to constructor must contain phenonaut Datasets or pandas DataFrames"
                )

            if ds_types[0] == Dataset:
                self.datasets.extend(dataset)
                return

            if ds_types[0] == pd.DataFrame:
                if isinstance(dataframe_name, str):
                    dataframe_name = [dataframe_name] * len(dataset)
                if dataframe_name is None:
                    dataframe_name = ["Dataset from pd.DataFrame"] * len(dataset)
                if len(dataframe_name) != len(dataset):
                    raise ValueError(
                        f"dataframe_name contained {len(dataframe_name)} name(s), but dataset contained {len(dataset)} pd.DataFrames"
                    )
                if metadata is None:
                    metadata = {}
                if isinstance(metadata, dict):
                    metadata = [metadata.copy() for dm in range(len(dataset))]
                if len(metadata) != len(dataset):
                    raise ValueError(
                        f"dataset_metadata contained {len(metadata)} metadata dictionaries, but dataset contained {len(dataset)} pd.DataFrames"
                    )
                for ds_name, df, ds_metadata in zip(dataframe_name, dataset, metadata):
                    self.datasets.append(Dataset(dataset_name=ds_name, input_file_path=df, metadata=ds_metadata))
                return

            if name is not None:
                self.name = name
            return
        # If a packaged dataset has been passed, then load it
        if issubclass(type(dataset), PackagedDataset):
            if isinstance(packaged_dataset_name_filter, str):
                packaged_dataset_name_filter = [packaged_dataset_name_filter]
            for key in dataset.keys():
                if packaged_dataset_name_filter is None:
                    self.datasets.append(dataset[key])
                else:
                    if key in packaged_dataset_name_filter:
                        self.datasets.append(dataset[key])
            if name is None:
                if dataset.name is not None:
                    self.name = dataset.name
            else:
                self.name = name
            return
        raise ValueError(
            f"Expected data to be a Dataset, pandas DataFrame, list of Datasets, list of pandas DataFrames, PackagedDataset, sklearn Bunch, or Phenonaut object, however, it was {type(dataset)}"
        )

    def __getitem__(self, dataset: Union[str, int, list[str], list[int]]) -> Dataset:
        """Allow dictionary-like access to Datasets

        Allows simple access to a Phenonaut object's Datasets using their index
        within phe.datasets, or by the name of the Dataset (instance.name)

        Parameters
        ----------
        dataset : Union[str, int]
            Can be the name of a dataset, or an index within Phenonaut.datasets.
            It can also be a list of Dataset names or indexes in order to return
            a tuple of requested Datasets.

        Returns
        -------
        Dataset
            The requested Dataset

        Raises
        ------
        KeyError
            Dataset with the supplied name was not found
        ValueError
            Invalid dataset index
        """
        if isinstance(dataset, str):
            if dataset not in self.keys():
                raise KeyError(
                    f"'{dataset}' not found as a name in datasets ({self.keys()})"
                )
            return self.datasets[self.keys().index(dataset)]
        elif isinstance(dataset, int):
            return self.datasets[dataset]
        elif isinstance(dataset, (list, tuple)):
            return [self[ds] for ds in dataset]
        else:
            raise ValueError(
                f"Attempting to get dataset, but {type(dataset)} was supplied. It should be str name, int giving the dataset location in the Phenonaut.datasets list, or a list/tuple of names/indexes to return multiple datasets"
            )

    def __setitem__(self, name: Union[str, int], dataset: Dataset) -> None:
        """Allows asignment of Datasets to Phenonaut object by name

        Parameters
        ----------
        name : str
            Name of dataset
        dataset : Dataset
            New Dataset to be stored in Phenonaut object
        """
        if isinstance(name, int):
            self.datasets[name]=dataset
        else:
            dataset.name = name
            if name in self.keys():
                name=self.get_dataset_index_from_name(name)
                self.datasets[name]=dataset
            else:
                self.datasets.append(dataset)

    def __delitem__(self, key: Union[str, int]):
        """Allows the deletion of Datasets stored in a Phenonaut object

        Allows the simple deletion of Datasets stored within a Phenonaut object
        by name or index.  Allows the simple notation:
        del phe['my_dataset']
        to delete a dataset called 'my_dataset', alternatively, we may use
        indexing like
        del phe[-1]
        to remove the last added Dataset to the Phenonaut object.

        Parameters
        ----------
        key : Union[str, int]
            Name of the dataset or the index of the Dataset to be removed.

        Raises
        ------
        KeyError
            Dataset of that name not found

        """
        if isinstance(key, str):
            if key not in self.keys():
                raise KeyError(
                    f"'{key}' not found as a name in datasets ({self.keys()})"
                )
            del self.datasets[self.keys().index(key)]
            return
        elif isinstance(key, int):
            del self.datasets[key]
            return

    def __iter__(self):
        """Like a dictionary, iterating a Phenonaut object returns dataset keys

        Returns
        -------
        List
            List of dataset names
        """
        return (k for k in self.keys())

    @property
    def ds(self) -> Dataset:
        """Return the dataset with the highest index in phenonaut.datasets

        Returns
        -------
        Dataset
            Last added/highest indexed Dataset

        Raises
        ------
        DataError
            No datasets loaded
        """
        if self.datasets is None or self.datasets == []:
            raise DataError("Attempt to get dataset, but no dataset loaded")
        return self.datasets[-1]

    @property
    def df(self) -> pd.DataFrame:
        """Return the pd.DataFrame of the last added/highest indexed Dataset

        Returns the internal pd.Dataframe of the Dataset contained within the
        Phenonaut instance's datasets list.

        Returns
        -------
        pd.DataFrame
            _description_
        """
        return self.ds.df

    def keys(self) -> list[str]:
        """Return a list of all dataset names

        Returns
        -------
        list(str)
            List of dataset names, empty list if no datasets are loaded.
        """
        if self.datasets is None or self.datasets == []:
            return []
        return [ds.name for ds in self.datasets]

    def get_hash_dictionary(self) -> dict:
        """Returns dictionary containing SHA256 hashes

        Returns a dictionary of base64 encoded UTF-8 strings representing the
        SHA256 hashes of datasets (along with names), combined datasets, and
        the Phenonaut object (including name).

        Returns
        -------
        dict
            Dictionary of base64 encoded SHA256 representing datasets and
            the Phenonaut object which created them.
        """
        ds_hashes = {
            f"{idx}:{ds.name}": b64encode(ds.sha256.digest()).decode("utf-8")
            for idx, ds in enumerate(self.datasets)
        }
        phenonaut_object_hash = sha256(self.init_hash)
        combined_ds_hash = sha256(self.init_hash)

        for ds in self.datasets:
            combined_ds_hash.update(ds.sha256.digest())
            phenonaut_object_hash.update(ds.sha256.digest())
        phenonaut_object_hash.update(self.name.encode("utf-8"))

        return {
            "datasets": ds_hashes,
            "combined_datasets_hash": b64encode(combined_ds_hash.digest()).decode("utf-8"),
            "phenonaut_object_name": self.name,
            "phenonaut_object_hash": b64encode(phenonaut_object_hash.digest()).decode("utf-8"),
        }

    def __delete__(self) -> None:
        """Deconstructor for Phenonaut object

        Upon deconstruction of a Phenonaut object, a hash dictionary is
        retrieved and printed to standard output.

        """
        print(
            "Phenonaut object destroyed, hash information\n"
            "--------------------------------------------\n"
        )
        for k, v in self.get_hash_dictionary().items():
            print(f"{k} : {v}")

    def load_dataset(
        self,
        dataset_name: str,
        input_file_path: Union[Path, str],
        metadata: dict = None,
        h5_key: Optional[str] = None,
        features:Optional[list[str]]=None,
    ):
        """Load a dataset from a CSV, optionally suppying metadata and a name

        Parameters
        ----------
        dataset_name : str
            Name to be assigned to the dataset
        input_file_path : Union[Path, str]
            CSV/TSV/H5 file location
        metadata : dict, optional
            Metadata dictionary describing the CSV data format, by default None
        h5_key : Optional[str]
            If input_file_path is an h5 file, then a key to access the target
            DataFrame must be supplied.
        features : Optional[list[str]]
            Optionally supply a list of features here. If None, then the
            features/feature finding related keys in metadata are used. You may
            also explicitly supply an empty list to explicitly specify that the
            dataset has no features. This is not recommended.
        """
        if dataset_name is None and metadata is not None:
            dataset_name = metadata.pop("dataset_name", f"Dataset {len(self.datasets)+1}")
        if features !=None:
            metadata['features']=features
        self.datasets.append(Dataset(dataset_name=dataset_name, input_file_path=input_file_path, metadata=metadata))

    def combine_datasets(self, dataset_ids_to_combine:Optional[Union[list[str], list[int]]]=None, new_name:Optional[str]=None, features: list = None):
        """Combine multiple datasets into a single dataset

        Often, large datasets are split across multiple CSV files. For example,
        one CSV file per screening plate.  In this instance, it is prudent to
        combine the datasets into one.

        Parameters
        ----------
        dataset_ids_to_combine : Optional[Union[list[str], list[int]]]
            List of dataset indexes, or list of names of datasets to combine.
            For example, after loading in 2 datasets, the list [0,1] would be given,
            or a list of their names resulting in a new third dataset in datasets[2].
            If None, then all present datasets are used for the merge. By default, None.
        new_name :new_name:Optional[str]
            Name that should be given to the newly created dataset. If None, then it is
            assigned as: "Combined_dataset from datasets[DS_INDEX_LIST]", where DS_INDEX_LIST
            is a list of combined dataset indexes.
        features : list, optional
            List of new features which should used by the newly created dataset
            if None, then features of combined datasets are used. By default
            None.

        Raises
        ------
        DataError
            Error raised if the combined datasets do not have the same features.
        """
        print(dataset_ids_to_combine)
        if dataset_ids_to_combine == []:
            raise ValueError("Empty list supplied as indexes of datasets to combine")
        if dataset_ids_to_combine is None:
            dataset_ids_to_combine=[idx for idx in range(len(self.datasets))]
        
        dataframes_to_combine = [self[i].df for i in dataset_ids_to_combine]
        datasets_to_combine = [self[i] for i in dataset_ids_to_combine]
        new_metadata = {}
        for ds in datasets_to_combine:
            for k, v in ds._metadata.items():
                new_metadata[k] = v
        new_df = pd.concat(dataframes_to_combine, axis=0).reset_index()

        if features is None:
            features = datasets_to_combine[0].features
            for ds in datasets_to_combine[1:]:
                if ds.features != features:
                    raise DataError(
                        f"Different features found when combining datasets\n{features}\nvs\n{ds.features}"
                    )
        if new_name is None:
            new_name = f"Combined_dataset from datasets[{dataset_ids_to_combine}]"
        self.datasets.append(
            Dataset(dataset_name=new_name,input_file_path=None, metadata=None)
        )
        self.datasets[-1].df = new_df
        self.datasets[-1]._metadata = new_metadata
        self.datasets[-1].features = (
            features,
            f"Combined datasets at dataset indexes: {dataset_ids_to_combine}",
        )

    def get_dataset_index_from_name(
        self, name: Union[str, list[str], tuple[str]]
    ) -> Union[int, list[int]]:
        """Get dataset index from name

        Given the name of a dataset, return the index of it in datasets list.
        Accepts single string query, or a list/tuple of names to return lists
        of indices.

        Parameters
        ----------
        name : Union[str, list[str], tuple[str]]
            If string, then this is the dataset name being searched for. Its
            index in the datasets list will be returned. If a list or tuple
            of names, then the index of each is searched and an index list
            returned.

        Returns
        -------
        Union[int, list[int]]
            If name argument is a string, then the dataset index is returned.
            If name argument is a list or tuple, then a list of indexes for
            each dataset name index is returned.

        Raises
        ------
        ValueError
            Error raised if no datasets were found to match a requested name.
        """
        dataset_names = [ds.name for ds in self.datasets]
        if isinstance(name, str):
            return dataset_names.index(name)
        elif isinstance(name, (list, tuple)):
            return [dataset_names.index(n) for n in name]

    def clone_dataset(
        self,
        existing_dataset: Union[Dataset, str, int],
        new_dataset_name: str,
        overwrite_existing: bool = False,
    ) -> None:
        """Clone a dataset into a new dataset

        Parameters
        ----------
        existing_dataset : Union[Dataset, str, int]
            The name or index of an existing Phenonaut Dataset held in the
            Phenonaut object. Can also be a Phenonaut.Dataset object passed
            directly.
        new_dataset_name : str
            A name for the new cloned Dataset.
        overwrite_existing : bool, optional
            If a dataset by this name exists, then overwrite it, otherwise, an
            exception is raised, by default False.

        Raises
        ------
        ValueError
            Dataset by the name given already exists and overwrite_existing was
            False.
        ValueError
            The existing_dataset argument should be a str, int or
            Phenonaut.Dataset.
        """
        if new_dataset_name in self.keys():
            if not overwrite_existing:
                raise ValueError(
                    f"new_dataset_name ({new_dataset_name}) is already the name of an "
                    "existing dataset, and overwrite_existing was False"
                )
        if isinstance(existing_dataset, (str, int)):
            new_ds = self[existing_dataset].copy()
        elif isinstance(existing_dataset, Dataset):
            new_ds = existing_dataset.copy()
        else:
            raise ValueError(
                f"The existing_dataset argument should be a string (giving the name of an "
                "existing dataset), an int (giving the index of an existing dataset), or "
                " a phenonaut.Dataset object"
            )
        new_ds.name = new_dataset_name

        if new_dataset_name in self.keys():
            self.datasets[self.keys().index(new_dataset_name)] = new_ds
        else:
            self.datasets.append(new_ds)

    def new_dataset_from_query(
        self,
        name,
        query: str,
        query_dataset_name_or_index: Union[int, str] = -1,
        raise_error_on_empty: bool = True,
        overwrite_existing: bool = False,
    ):
        """Add new dataset through a pandas query of existing dataset

        Parameters
        ----------
        query : str
            The pandas query used to select the new dataset
        name : str
            A name for the new dataset
        query_dataset_name_or_index : Union[int, str], optional
            The dataset to be queried, can be an int index, or the name of an
            existing dataset. List indexing can also be used, such that -1 uses
            the last dataset in Phenonaut.datasets list, by default -1.
        raise_error_on_empty : bool
            Raise a ValueError is the query returns an empty dataset.  By
            default True.
        overwrite_existing : bool
            If a dataset already exists with the name given in the name
            argument, then this argument can be used to overwrite it, by
            default False.
        """
        new_ds = self[query_dataset_name_or_index].get_ds_from_query(name, query)
        if raise_error_on_empty and new_ds.df.shape[0] == 0:
            raise ValueError("Query used to create a new dataset returned no results")
        if name in self.keys():
            if overwrite_existing:
                self[name] = new_ds
            else:
                raise ValueError(
                    "Dataset with name given in the name argument already exists, "
                    "and the overwrite_existing argument was False"
                )
        else:
            self[name] = new_ds

    def add_well_id(
        self,
        numerical_column_name: str = "COLUMN",
        numerical_row_name: str = "ROW",
        plate_type: int = 384,
        new_well_column_name: str = "Well",
        add_empty_wells: bool = False,
        plate_barcode_column: str = None,
        no_sort: bool = False,
    ):
        """Add standard well IDs - such as A1, A2, etc to ALL loaded Datasets.

        If a dataset contains numerical row and column names, then they may be
        translated into standard letter-number well IDs. This is applied to all
        loaded Datasets.  If you wish only one to be annotated, then call
        add_well_id on that individual dataset.

        Parameters
        ----------
        numerical_column_name : str, optional
            Name of column containing numeric column number, by default
            "COLUMN".
        numerical_row_name : str, optional
            Name of column containing numeric column number, by default "ROW".
        plate_type : int, optional
            Plate type - note, at present, only 384 well plate format is
            supported, by default 384.
        new_well_column_name : str, optional
            Name of new column containing letter-number well ID, by default
            "Well".
        add_empty_wells : bool, optional
            Should all wells from a plate be inserted, even when missing from
            the data, by default False.
        plate_barcode_column : str, optional
            Multiple plates may be in a dataset, this column contains their
            unique ID, by default None.
        no_sort : bool, optional
            Do not resort the dataset by well ID, by default False
        """
        for ds in self.datasets:
            ds.add_well_id(
                numerical_column_name=numerical_column_name,
                numerical_row_name=numerical_row_name,
                plate_type=plate_type,
                new_well_column_name=new_well_column_name,
                add_empty_wells=add_empty_wells,
                plate_barcode_column=plate_barcode_column,
                no_sort=no_sort,
            )

    def subtract_median_perturbation(
        self,
        perturbation_label: str,
        per_column_name: Optional[str] = None,
        new_features_prefix: str = "SMP_",
    ):
        """Subtract the median perturbation from all features for all datasets.

        Useful for normalisation within a well/plate format. The median feature
        may be identified through the per_column_name variable, and perturbation
        label. Newly generated features may have their prefixes controled via
        the new_features_prefix argument.

        Parameters
        ----------
        perturbation_label : str
            The perturbation label which should be used to calculate the median
        per_column_name : Optional[str], optional
            The perturbation column name.  This is optional and can be None, as
            the Dataset may already have perturbation column set. By
            default, None.
        new_features_prefix : str
            Prefix for new features, each with the median perturbation
            subtracted. By default 'SMP\_' (for subtracted median perturbation).

        """
        for ds in self.datasets:
            ds.subtract_median_perturbation(
                perturbation_label,
                per_column_name,
                new_features_prefix,
            )

    def get_dataset_combinations(
        self,
        min_datasets: Optional[int] = None,
        max_datasets: Optional[int] = None,
        return_indexes: bool = False,
    ):
        """Get tuple of all dataset name combinations, picking 1 to n datasets

        The function to return all combinations from 1 to n dataset names in
        combination, where n is the number of loaded datasets. This is useful
        in multiomics settings where we test A,B, and C alone, A&B, A&C, B&C,
        and finally A&B&C.

        A limit on the number of datasets in a combination can be imposed
        using the max_datasets argument. In the example above with datasets
        A, B and C, passing max_datasets=2 would return the following tuple:
        ((A), (B), (C), (A, B), (A, C), (B, C))
        leaving out the tripple length combination (A, B, C).

        Similarly, the argument min_datasets can specify a lower limit on
        the number of dataset combinations.

        Using the example with datasets A, B, and C, and setting min_datasets=2
        with no limit on max_datasets on the above example would return the
        following tuple:
        ((A, B), (A, C), (B, C), (A, B, C))

        If return_indexes is True, then the indexes of Datasets are returned.
        As directly above, datasets A, B, and C, setting min_datasets=2
        with no limit on max_datasets and passing return_indexes=True would
        return the following tuple:
        ((0, 1), (0, 2), (1, 2), (0, 1, 2))

        Parameters
        ----------
        min_datasets : Optional[int], optional
            Minimum number of datasets to return in a combination. If None,
            then it behaves as if 1 is given, by default None.
        max_datasets : Optional[int], optional
            Maximum number of datasets to return in a combination. If None,
            then it behaves as if len(datasets) is given, by default None.
        return_indexes : bool
            Return indexes of Datasets, instead of their names,
            by default False.
        """
        ds_names = None
        if return_indexes:
            ds_names = range(len(self.datasets))
        else:
            ds_names = self.keys()

        if max_datasets is None:
            max_datasets = len(ds_names)
        if min_datasets is None:
            min_datasets = 1
        if min_datasets > max_datasets:
            raise ValueError("min_datasets cannot be larger than max_datasets")

        return tuple(
            c
            for i in range(min_datasets, max_datasets + 1)
            for c in itertools_combinations(ds_names, i)
        )


    def aggregate_dataset(
        self,
        composite_identifier_columns: list[str],
        datasets: Union[Iterable[int], Iterable[str], int, str] = -1,
        new_names_or_prefix: Union[list[str], tuple[str], str] = "Aggregated_",
        inplace: bool = False,
        transformation_lookup: dict[str, Union[Callable, str]] = None,
        tranformation_lookup_default_value: Union[str, Callable] = "mean",
    ):
        """Aggregate multiple or single phenonaut dataset rows

        If we have a Phenonaut object containing data derived from 2 fields of
        view from a microscopy image, a sensible approach is averaging features.
        If we have the DataFrame below, we may merge FOV 1 and FOV 2, taking the
        mean of all features.  As strings such as filenames should be kept, they
        are concatenated together, separated by a comma, unless the strings are
        the same, in which case just one is used.

        Here we test a df as follows:

        === ====== ======= ====== ====== ====== ========= ===
        ROW COLUMN BARCODE feat_1 feat_2 feat_3 filename  FOV
        === ====== ======= ====== ====== ====== ========= ===
        1   1      Plate1  1.2    1.2    1.3    FileA.png 1
        1   1      Plate1  1.3    1.4    1.5    FileB.png 2
        1   1      Plate2  5.2    5.1    5      FileC.png 1
        1   1      Plate2  6.2    6.1    6.8    FileD.png 2
        1   2      Plate1  0.1    0.2    0.3    FileE.png 1
        1   2      Plate1  0.2    0.2    0.38   FileF.png 2
        === ====== ======= ====== ====== ====== ========= ===

        With just this loaded into a phenonaut object, we can call:

        phe.aggregate_dataset(['ROW','COLUMN','BARCODE'])

        Will merge and proeduce another, secondary dataset in the phe object
        containing:

        === ====== ======= ======= ======= =======  =================== ====
        ROW COLUMN BARCODE  feat_1  feat_2  feat_3            filename   FOV
        === ====== ======= ======= ======= =======  =================== ====
        1       1   Plate1    1.25     1.3    1.40  fileA.png,FileB.png  1.5
        1       1   Plate2    5.70     5.6    5.90  FileC.png,FileD.png  1.5
        1       2   Plate1    0.15     0.2    0.34  FileF.png,fileE.png  1.5
        === ====== ======= ======= ======= =======  =================== ====

        if inplace=True is passed in the call to aggregate_dataset, then the
        phenonaut object will contain just one dataset, the new aggregated
        dataset.

        Parameters
        ----------
        composite_identifier_columns : list[str]
            If a biochemical assay evaluated through imaging is identified by a
            row, column, and barcode (for the plate) but multiple images taken
            from a well, then these multiple fields of view can be merged,
            creating averaged features using row, column and barcode as the
            composite identifier on which to merge fields of view.
        datasets : Union[list[int], list[str], int, str]
            Which datasets to apply the aggregation to.  If int, then the
            dataset with that index undergoes aggregation. If a string, then the
            dataset with that name undergoes aggregation. It may also be a list
            or tuple of mixed int and string types, with ints specifying dataset
            indexes and strings indicating dataset names. By default, this value is -1,
            indicating that the last added dataset should undergo aggregation.
        new_names_or_prefix : Union[list[str], tuple[str], str]
            If a list or tuple of strings is passed, then use them as the names
            for the new datasets after aggregation. If a single string is
            passed, then use this as a prefix for the new dataset.
            By default "Aggregated\_".
        inplace : bool
            Perform the aggregation in place, overwriting the origianl
            dataframes. By default False.
        transformation_lookup : dict[str,Union[Callable, str]]
            Dictionary mapping data types to aggregations. When None, it is as
            if the dictionary:
            {np.dtype("O"): lambda x: ",".join([f"{item}" for item in set(x)])}
            was provided, concatenating strings together (separated by a comma)
            if they are different and just using one if they are the same across
            rows. If a type not present in the dictionary is encountered (such
            as int, or float) in the above example, then the default specified
            by transformation_lookup_default_value is returned.
            By default, None.
        tranformation_lookup_default_value : Union[str, Callable]
            Transformation to apply if the data type is not found in the
            transformation_lookup_dictionary, can be a callable or string to
            pandas defined string to function shortcut mappings.
            By default "mean".

        """
        
        if isinstance(datasets, (int, str)):
            datasets = [datasets]
        datasets=list(datasets) # In case generator is consumed by new_names_or_prefix below.

        if isinstance(new_names_or_prefix, str):
            new_names_or_prefix = [
                f"{new_names_or_prefix}{self[idx].name}"
                for idx in datasets
            ]

        if len(new_names_or_prefix) != len(datasets):
            raise ValueError(
                f"Expected new_names_or_prefix to have the same number of elements as number datasets being worked on. Was {len(new_names_or_prefix)}, when it should be ({len(datasets)})"
            )
        for dsi, new_ds_name in zip(datasets, new_names_or_prefix):
            new_ds = self[dsi].new_aggregated_dataset(
                composite_identifier_columns,
                new_dataset_name=new_ds_name,
                transformation_lookup=transformation_lookup,
                tranformation_lookup_default_value=tranformation_lookup_default_value,
            )
            if inplace:
                self[dsi] = new_ds
            else:
                self.datasets.append(new_ds)

    def merge_datasets(
        self,
        ids: Optional[Union[list, tuple]] = None,
        common_merge_columns: Optional[list[str]] = None,
        new_dataset_name: str = "merged_dataset",
    ):
        """Merge Datasets within a Phenonaut object

        With multiple views of an underlying biological system, it is often
        useful to merge different datasets, for example, DRUG-seq with
        CellProfiler or similar imaging data. This function merges Datasets
        and adds the resultant Dataset to the Phenonaut object.  DataFrames are
        assumed to have common columns which allow the merge to proceed.

        Underneath the hood, Phenonaut uses Panda's pd.merge function, and then
        handles features and other Phenonaut specific data like features.

        Parameters
        ----------
        ids : Optional[Union[list, tuple]], optional
            A list or tuple containing two indexes, or dataset names contained
            within the Phenonaut object which should be merged. If None, then
            it is assumed that the Phenonaut object contains only 2 datasets,
            and their indexes are then used.  If more than 2 datasets are found
            and ids is None, then an error is raised. By default None.
        common_merge_columns : Optional[list[str]], optional
            Merge datasets on common columns, found within each dataset. This may
            be things like a treatment ID, WellID, PlateID, etc. If None, then
            both Datasets are looked at and common columns identified. By
            default None.
        new_dataset_name : str, optional
            Name for the new dataset, by default "merged_dataset".

        Raises
        ------
        ValueError
            The ids argument was None, but more than 2 datasets were found in
            the Phenonaut object, please specify which Datasets should be
            merged.
        ValueError
            The ids argument contained more than 2 elements.
        ValueError
            Datasets contain common feature columns, cannot merge.
        ValueError
            No common columns found to perform merge.
            
        """        
        if ids is None:
            ids = self.get_dataset_index_from_name(self.keys())
            if len(ids) != 2:
                raise ValueError(
                    f"merge_datasets merges 2 Datasets at a time.  ids was None, and there are {len(ids)} datasets in this Phenonaut object"
                )
        if len(ids) != 2:
            raise ValueError(
                f"merge_datasets merges 2 Datasets at a time. The ids argument contained {len(ids)} dataset indexes"
            )
        if any(f1 in self[ids[1]].features for f1 in self[ids[0]].features):
            raise ValueError(f"Datasets shared common features, cannot merge, common features were {set(self[ids[0]].features).intersection(set(self[ids[1]].features))}")
        if common_merge_columns is None:
            common_merge_columns = list(
                set.intersection(*map(set, [self[id].df.columns.tolist() for id in ids]))
            )
        if len(common_merge_columns) == 0:
            raise ValueError("No common columns found in dataframes to perform merge")
        new_df = pd.merge(self[ids[0]].df, self[ids[1]].df, on=common_merge_columns)
        # Uniquify features, preserving order
        merged_features=self[ids[0]].features+self[ids[1]].features

        self.datasets.append(Dataset(dataset_name=new_dataset_name,input_file_path=new_df, metadata={'features':merged_features}))

    def save(self, output_filename: Union[str, Path], overwrite_existing: bool = False) -> None:
        """Save Phenonaut object and contained Data to a pickle

        Writes a gzipped Python pickle file. If no compression, or another
        compressison format is required, then the user should use a custom
        pickle.dump and not rely on this helper function.  Warning, SHA256
        '_hashlib.HASH' objects are not serialisable, so crytographic hashes
        and proof of manipulations are lost upon saving.

        Parameters
        ----------
        output_filename : Union[str, Path]
            Output filename for the gzipped pickle
        overwrite_existing : bool, optional
            If True and the file exists, overwrite it. By default False.
        """
        for ds in self.datasets:
            ds.sha256 = b64encode(ds.sha256.digest()).decode("utf-8")
        output_filename = check_path(output_filename)
        with gzip.open(output_filename, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filepath: Union[str, Path]) -> "Phenonaut":
        """Class method to load a compressed Phenonaut object

        Loads a gzipped Python pickle containing a Phenonaut object

        Parameters
        ----------
        filepath : Union[str, Path]
            Location of gzipped Phenonaut object pickle

        Returns
        -------
        phenonaut.Phenonaut
            Loaded Phenonaut object.

        Raises
        ------
        FileNotFoundError
            File not found, unable to load pickled Phenonaut object.
        """
        filepath = check_path(filepath, make_parents=False)
        if not filepath.exists():
            raise FileNotFoundError(f"Could not find file {filepath}")
        with gzip.open(filepath, "rb") as f:
            phe = pickle.load(f)
            for ds in phe.datasets:
                ds.sha256 = sha256(ds.sha256.encode("utf-8"))
            return phe



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from multiprocessing.sharedctypes import Value
from sklearn import decomposition
from phenonaut import data
from phenonaut.data.dataset import Dataset
from typing import Callable, Iterable, Optional, Type, Union, List
from inspect import isclass
from copy import deepcopy
from numpy import array
import numpy as np
import pandas as pd

from phenonaut.phenonaut import Phenonaut


class Transformer:
    """Generic transformer to turn methods/objects into Phenonaut transforms

    This generic transformer class may be used to wrap functions which perform
    a simple transform (like np.log2), or even more complex objects, like PCA,
    t-SNE and StandardScaler from scikit (which provide their own fit/transform/
    fit_transform functions).

    This way, a Phenonaut transformer may be constructed with applies the given
    function/object method to a Phenonaut dataset and correctly updates features.

    Wrapping an object which requires fitting also brings with it the advantage
    that we may apply the group_by keyword to perform a unique fit and transform
    in groups, which is useful if you require PCA to be performed on a per-plate
    basis.

    May be used as follows - in this example, we are wrapping PCA from SciKit
    which has the effect of perfoming a 2D PCA on our Phenonaut Dataset object.
    We also make use of the group_by keyword, to perform a unique PCA for each
    plate (denoted by unique BARCODE values).


    .. code-block:: python

        from phenonaut import Phenonaut
        from phenonaut.transforms import Transformer
        import pandas as pd
        import numpy as np

        df=pd.DataFrame({
            'ROW':[1,1,1,1,1,1],
            'COLUMN':[1,1,1,2,2,2],
            'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
            'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
            'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
            'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
            'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
            'FOV':[1,2,1,2,1,2]})

        phe=Phenonaut(df)
        from sklearn.decomposition import PCA
        t_pca=Transformer(PCA, constructor_kwargs={'n_components':2})
        t_pca.fit(phe.ds, group_by="BARCODE")
        t_pca.transform(phe.ds)

    Along with the above, whereby a PCA transformer is generated from the SciKit
    PCA class, you may also use the built in phenonaut.transforms.PCA which has
    more PCA specific functionality, allowing creation of Scree plots etc.

    A transformer may also be made with a callable function (which operates on
    dataframes or numpy arrays) like np.log2, or np.abs, etc, as shown below.

    The following squares all features.

    .. code-block:: python

        from phenonaut import Phenonaut
        from phenonaut.transforms import Transformer
        import numpy as np
        import pandas as pd
        df=pd.DataFrame({
            'ROW':[1,1,1,1,1,1],
            'COLUMN':[1,1,1,2,2,2],
            'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
            'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
            'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
            'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
            'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
            'FOV':[1,2,1,2,1,2]})
        phe=Phenonaut(df)
        t=Transformer(np.square)
        t(phe)



    Parameters
    ----------
    method : Union[object, Callable]
        Instantiatable object with fit, transform and or fit_transform, or
        __call__ methods. Alternatively, a callable function. Designed to be
        passed something like the PCA class from SciKit, or a simple function
        like np.log2.
    new_feature_names : Union[list[str], str]
        List of strings containing the names for the new features. Can also
        be just a single string, which then has numerical suffixes attached
        enumerating the number of new features generated.
    transformer_name : Optional[str], optional
        When features are set on a Dataset, the history reflects what was
        carried out to generate those features. If None, then automatic naming
        of the passed function is attempted. By default None.
    constructor_kwargs : dict, optional
        Additional constructor arguments which may be passed to the class passed
        in the method argument upon instantiation. By default {}.
    callable_kwargs : dict, optional
        Additional arguments to pass to the function passed in the method
        argument. By default {}.
    fit_kwargs : dict, optional
        Additional arguments to the fit function called on object instantiated
        from the method argument. By default {}.
    transform_kwargs : dict, optional
        Additional arguments to the transform function called on object
        instantiated from the method argument. By default {}.
    fit_transform_kwargs : dict, optional
        Additional arguments to the fit_transform function called on object
        instantiated from the method argument. By default {}.
    """

    def __init__(
        self,
        method: Union[object, Callable],
        new_feature_names: Optional[Union[str, list[str]]] = None,
        transformer_name: Optional[str] = None,
        constructor_kwargs: dict = {},
        callable_kwargs: dict = {},
        fit_kwargs: dict = {},
        transform_kwargs: dict = {},
        fit_transform_kwargs: dict = {},
    ):

        self._original_method = method
        self._new_feature_names = new_feature_names
        self._constructor_kwargs = constructor_kwargs
        self._callable_kwargs = callable_kwargs
        self._fit_kwargs = fit_kwargs
        self._transform_kwargs = transform_kwargs
        self._fit_transform_kwargs = fit_transform_kwargs
        self._tranformer_name = transformer_name

        # Here we need to handle a method being put in, a class, and a class instance.
        if isclass(method):
            method = method(**constructor_kwargs)
        self._method = method

        self._has_fit = True if hasattr(self._method, "fit") else False
        self._has_transform = True if hasattr(self._method, "transform") else False
        self._has_fit_transform = True if hasattr(self._method, "fit_transform") else False
        self._is_callable = True if hasattr(self._method, "__call__") else False

    def fit(
        self,
        dataset: Union[Dataset, Phenonaut],
        group_by: Optional[Union[str, list[str]]] = None,
        fit_perturbation_ids: Union[str, list] = None,
        fit_query: Optional[str] = None,
        fit_kwargs: Optional[dict] = None,
    ):
        """Call fit on the transformer

        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            Dataset containing data to be fitted against
        group_by : Optional[Union[str, list[str]]], optional
            Often we would like to apply transformations on a plate-by-plate
            basis. This group_by argument allows definition of a string which
            is used to identify columns with which unique values define a group
            or plate. It works with the pandas group_by function and accepts
            a list of strings if multiple columns must be used to define groups.
            By default None.
        fit_perturbation_ids : Union[str, list], optional
            If only a subset of the data should be used for fitting, then
            pertubations for fitting may be listed here. If none, then
            every datapoint is used for fitting, by default None.
        fit_query : Optional[str], optional
            A pandas style query may be supplied to perform fitting. By default
            None.
        fit_kwargs : Optional[dict], optional
            Optional arguments supplied to the fit function of the object passed
            in the constructor of this transformer. By default None.

        """
        if fit_kwargs is None:
            fit_kwargs = self._fit_kwargs
        if fit_perturbation_ids is not None and fit_query is not None:
            raise ValueError(
                "Cannot supply both fit_perturbation_ids and fit_query together, only one can be used"
            )
        if not isinstance(dataset, (Phenonaut, Dataset)):
            raise TypeError(
                "Supplied data must be of type phenonaut.Dataset, or a Phenonaut object, "
                "in which case behavior is as if .ds was called to return the dataset with"
                f" the highest index (last added), found type was {type(dataset)}"
            )

        if isinstance(dataset, Phenonaut):
            dataset = dataset.ds

        if fit_perturbation_ids is not None:
            if isinstance(fit_perturbation_ids, str):
                fit_perturbation_ids = [fit_perturbation_ids]
            fit_perturbation_ids = list(fit_perturbation_ids)
            fit_query = f"{dataset.perturbation_column} == @fit_perturbation_ids"

        if group_by is None:
            if fit_query is None:
                self._method.fit(dataset.data, **fit_kwargs)
            else:
                self._method.fit(dataset.df.query(fit_query)[dataset.features], **fit_kwargs)
        else:
            self._original_features = dataset.features
            self._grouped_data = dataset.df.groupby(group_by)
            self._dataframes = [gdf for _, gdf in self._grouped_data]
            self._method = []
            for df in self._dataframes:
                if isclass(self._original_method):
                    self._method.append(self._original_method(**self._constructor_kwargs))
                else:
                    self._method.append(deepcopy(self._orignal_method))
                if fit_query is None:
                    self._method[-1].fit(df[self._original_features], **fit_kwargs)
                else:
                    self._method[-1].fit(df.query(fit_query)[self._original_features], **fit_kwargs)
        return self

    def transform(
        self,
        dataset: Union[Dataset, Phenonaut],
        new_feature_names: Optional[Union[list[str], str]] = None,
        method_name: Optional[str] = None,
        transform_kwargs: Optional[dict] = None,
        free_memory_after_transform: bool = True,
        center_on_perturbation_id: Optional[str] = False,
        centering_function: Callable = np.median,
    ):
        """Apply transform

        If no transform method is found, then __call__ is called.

        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object upon which the transform
            should be applied.
        new_feature_names : Optional[Union[list[str], str]], optional
            List of strings containing the names for the new features. Can also
            be just a single string, which then has numerical suffixes attached
            enumerating the number of new features generated. If None, then
            features are attempted to be named through interrogation of the
            method_name argument. By default None.
        method_name : Optional[str], optional
            When setting features, the history message should describe what was
            done. If None, then the method object/method is interrogated in an
            attempt to automatically deduce the name. By default None.
        transform_kwargs : Optional[dict], optional
            Additional arguments to the transform function called on object
            instantiated from the method argument. If set here, and not None, then
            it overrides any transform_kwargs given to the object constructor. By
            default None.
        free_memory_after_transform : bool, optional
            Remove temporary objects once transformation has been performed.
            By default True.
        center_on_perturbation_id : Optional[str], optional
            Optionally, recentre the PCA on a named perturbation. Should have
            pertubation_column set within the dataset for this option, by
            default None.
        centering_function : Callable, optional
            Used with center_on_perturbation, this function is applied to
            all data for matching perturbations. By default, we use the median
            of perturbations. This behavior can be overridden by supplying a
            different function here. By default np.median.

        """
        if not self._has_transform:
            if self._is_callable:
                self.__call__(dataset, new_feature_names=new_feature_names, method_name=method_name)
            else:
                raise ValueError("No transform method found, and transform is also not callable.")
        if transform_kwargs is None:
            transform_kwargs = self._transform_kwargs
        if not isinstance(dataset, (Phenonaut, Dataset)):
            raise TypeError(
                "Supplied data must be of type phenonaut.Dataset, or a Phenonaut object, "
                "in which case behavior is as if .ds was called to return the dataset with"
                f" the highest index (last added), found type was {type(dataset)}"
            )
        if isinstance(dataset, Phenonaut):
            dataset = dataset.ds

        if isinstance(self._method, list):
            if len(self._method) == 0 or len(self._method) != len(self._dataframes):
                raise ValueError(
                    f"groupby appears to have failed, there are {len(self._method)} transformers, and {len(self._dataframes)} groups"
                )

            for df_index, (fitted_object, df) in enumerate(zip(self._method, self._dataframes)):
                transformed_data_array = np.array(
                    fitted_object.transform(df[self._original_features], **transform_kwargs)
                )
                method_name, new_feature_names = self._get_method_name_and_feature_names(
                    transformed_data_array.shape[1], method_name, new_feature_names
                )

                transformed_df = pd.concat(
                    [
                        df,
                        pd.DataFrame(
                            data=transformed_data_array, columns=new_feature_names, index=df.index
                        ),
                    ],
                    axis=1,
                )

                self._dataframes[df_index] = self._center_df(
                    dataset,
                    transformed_df,
                    center_on_perturbation_id,
                    new_feature_names,
                    centering_function,
                )

            dataset.df = pd.concat(self._dataframes, axis=0, ignore_index=True)
            if free_memory_after_transform:
                del self._dataframes
                del transformed_data_array
                self._method = (
                    self._original_method(**self._constructor_kwargs)
                    if isclass(self._original_method)
                    else self._original_method
                )
        else:
            transformed_data_array = np.array(
                self._method.transform(dataset.data, **transform_kwargs)
            )

            method_name, new_feature_names = self._get_method_name_and_feature_names(
                transformed_data_array.shape[1], method_name, new_feature_names
            )
            transformed_df = pd.concat(
                [
                    dataset.df,
                    pd.DataFrame(
                        data=transformed_data_array,
                        columns=new_feature_names,
                        index=dataset.df.index,
                    ),
                ],
                axis=1,
            )
            # Attempt to center on perturbation ids,
            dataset.df = self._center_df(
                dataset,
                transformed_df,
                center_on_perturbation_id,
                new_feature_names,
                centering_function,
            )

        dataset.features = (
            new_feature_names,
            f"Applied {method_name}, {self._constructor_kwargs=}, {self._fit_kwargs=}, {self._transform_kwargs=}, {self._fit_transform_kwargs=}, {self._callable_kwargs=}, centered on {center_on_perturbation_id}, {centering_function}",
        )

    def _center_df(
        self, dataset, df, center_on_perturbation_id, new_feature_names, centering_function
    ):
        if center_on_perturbation_id is not None:
            if isinstance(center_on_perturbation_id, str):
                center_on_coords = centering_function(
                    df.query(f"{dataset.perturbation_column} =='{center_on_perturbation_id}'")[
                        new_feature_names
                    ],
                    axis=0,
                )
            else:
                center_on_coords = center_on_perturbation_id
            df.loc[:, new_feature_names] = df.loc[:, new_feature_names] - center_on_coords
        return df

    def fit_transform(
        self,
        dataset: Union[Dataset, Phenonaut],
        group_by: Optional[Union[str, list[str]]] = None,
        fit_perturbation_ids: Union[str, list] = None,
        fit_query: Optional[str] = None,
        fit_kwargs: Optional[dict] = None,
        transform_kwargs: Optional[dict] = None,
        fit_transform_kwargs: Optional[dict] = None,
        new_feature_names: Optional[Union[str, list[str]]] = None,
        method_name: Optional[str] = None,
        free_memory_after_transform: bool = True,
        center_on_perturbation_id: Optional[str] = False,
        centering_function: Callable = np.mean,
    ):
        """_summary_

        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object upon which the
            fit_transform should be applied.
        group_by : Optional[Union[str, list[str]]], optional
            Often we would like to apply transformations on a plate-by-plate
            basis. This group_by argument allows definition of a string which
            is used to identify columns with which unique values define a group
            or plate. It works with the pandas group_by function and accepts
            a list of strings if multiple columns must be used to define groups.
            By default None.
        fit_perturbation_ids : Union[str, list], optional
            If only a subset of the data should be used for fitting, then
            pertubations for fitting may be listed here. If none, then
            every datapoint is used for fitting, by default None.
        fit_query : Optional[str], optional
            A pandas style query may be supplied to perform fitting. By default
            None.
        fit_kwargs : Optional[dict], optional
            Optional arguments supplied to the fit function of the object passed
            in the constructor of this transformer. By default None.
        transform_kwargs : Optional[dict], optional
            Additional arguments to the transform function called on object
            instantiated from the method argument. If set here, and not None, then
            it overrides any transform_kwargs given to the object constructor.
            Transform is only called as a fallback, if the object does not have
            a fit_transform method, but separate fit and transform methods which
            may be called in series. By default None.
        fit_transform_kwargs : Optional[dict], optional
            Additional arguments to the fit_transform function called on object
            instantiated from the method argument. If set here, and not None, then
            it overrides any fit_transform_kwargs given to the object constructor.
            By default None.
        new_feature_names : Optional[Union[list[str], str]], optional
            List of strings containing the names for the new features. Can also
            be just a single string, which then has numerical suffixes attached
            enumerating the number of new features generated. If None, then
            features are attempted to be named through interrogation of the
            method_name argument. By default None.
        method_name : Optional[str], optional
            When setting features, the history message should describe what was
            done. If None, then the method object/method is interrogated in an
            attempt to automatically deduce the name. By default None.
        free_memory_after_transform : bool, optional
            Remove temporary objects once transformation has been performed.
            By default True.
        center_on_perturbation_id : Optional[str], optional
            Optionally, recentre the PCA on a named perturbation. Should have
            pertubation_column set within the dataset for this option, by
            default None.
        centering_function : Callable, optional
            Used with center_on_perturbation, this function is applied to
            all data for matching perturbations. By default, we use the median
            of perturbations. This behavior can be overridden by supplying a
            different function here. By default np.median.

        """
        if not isinstance(dataset, (Phenonaut, Dataset)):
            raise TypeError(
                f"Dataset should be a Phenonaut object, or phenonaut.Dataset object, it was {type(dataset)}"
            )
        if isinstance(dataset, Phenonaut):
            dataset = dataset.ds

        if fit_kwargs is None:
            fit_kwargs = self._fit_kwargs
        if transform_kwargs is None:
            transform_kwargs = self._fit_kwargs
        if fit_transform_kwargs is None:
            fit_transform_kwargs = self._fit_transform_kwargs

        if not self._has_fit_transform:
            self.fit(
                dataset,
                group_by=group_by,
                fit_perturbation_ids=fit_perturbation_ids,
                fit_query=fit_query,
                fit_kwargs=fit_kwargs,
            )
            self.transform(
                dataset,
                new_feature_names=new_feature_names,
                transform_kwargs=transform_kwargs,
                free_memory_after_transform=free_memory_after_transform,
                center_on_perturbation_id=center_on_perturbation_id,
                centering_function=centering_function,
            )
        else:
            if fit_perturbation_ids is not None or fit_query is not None:
                raise ValueError(
                    f"fit_perturbation_ids and fit_query cannot be used when the transformation has a native fit_transform method.  Call fit, then transform on the transformer"
                )
            if group_by is None:
                # Below we check if fit_transform_kwargs has anythin in it. This is not needed
                # for scikit and similar, but the UMAP package thinks that a y argument of length 0
                # if being passed if anything is there, so we work around it.
                if fit_transform_kwargs is None or fit_transform_kwargs == {}:
                    transformed_data_array = np.array(self._method.fit_transform(dataset.data))
                else:
                    transformed_data_array = np.array(
                        self._method.fit_transform(dataset.data, **fit_transform_kwargs)
                    )
                method_name, new_feature_names = self._get_method_name_and_feature_names(
                    transformed_data_array.shape[1], method_name, new_feature_names
                )
                transformed_df = pd.concat(
                    [
                        dataset.df,
                        pd.DataFrame(
                            data=transformed_data_array,
                            columns=new_feature_names,
                            index=dataset.df.index,
                        ),
                    ],
                    axis=1,
                )
                dataset.df = self._center_df(
                    dataset,
                    transformed_df,
                    center_on_perturbation_id,
                    new_feature_names,
                    centering_function,
                )

            else:  # Groupby is used
                self._original_features = dataset.features
                self._grouped_data = dataset.df.groupby(group_by)
                self._dataframes = [gdf for _, gdf in self._grouped_data]
                self._method = []
                for df_index, df in enumerate(self._dataframes):
                    if isclass(self._original_method):
                        self._method.append(self._original_method(**self._constructor_kwargs))
                    else:
                        self._method.append(deepcopy(self._orignal_method))
                    transformed_data_array = np.array(
                        self._method[-1].fit_transform(df[dataset.features], **fit_transform_kwargs)
                    )

                    method_name, new_feature_names = self._get_method_name_and_feature_names(
                        transformed_data_array.shape[1], method_name, new_feature_names
                    )

                    transformed_df = pd.concat(
                        [
                            df,
                            pd.DataFrame(
                                data=transformed_data_array,
                                columns=new_feature_names,
                                index=df.index,
                            ),
                        ],
                        axis=1,
                    )
                    self._dataframes[df_index] = self._center_df(
                        dataset,
                        transformed_df,
                        center_on_perturbation_id,
                        new_feature_names,
                        centering_function,
                    )
                dataset.df = pd.concat(self._dataframes, axis=0, ignore_index=True)
                if free_memory_after_transform:
                    del self._dataframes
            dataset.features = (
                new_feature_names,
                f"Applied {method_name}, {self._constructor_kwargs=}, {self._fit_kwargs=}, {self._transform_kwargs=}, {self._fit_transform_kwargs=}, {self._callable_kwargs=}",
            )
            if free_memory_after_transform:
                del transformed_data_array

    def _get_method_name_and_feature_names(
        self,
        n_features: int,
        method_name: Optional[str] = None,
        features: Optional[list[str]] = None,
    ):
        """Get method name and new feature names

        Parameters
        ----------
        n_features : int
            Number of new features to generate
        method_name : Optional[str], optional
            Name of the method, if known. If supplied and not None, then this
            name will be used. By default None.
        features : Optional[list[str]], optional
            If the new features are known, then these are returned instead, by default None

        Returns
        -------
        tuple(str, list[str])
            Tuple containing the method name and new feature names list.
        """
        if method_name is None:
            if isinstance(self._method, list):
                m = self._method[-1]
            else:
                m = self._method
            if hasattr(m, "__name__"):
                method_name = m.__name__
            elif hasattr(m, "__class__"):
                method_name = m.__class__.__name__
            else:
                method_name = "UnknownTransform"

        if features is None:
            features = self._new_feature_names
        if features is None:
            features = method_name
        if isinstance(features, str):
            features = [f"{features}_{i+1}" for i in range(n_features)]
        return method_name, features

    def __call__(
        self,
        dataset: Union[Dataset, Phenonaut],
        new_feature_names: Optional[Union[list[str], str]] = None,
        method_name: Optional[str] = None,
    ):
        """Call transformer

        If a simple callable method was passed to the constructor of the transformer,
        then it can be applied by calling the method here.  The transform method can
        also be called with the same effect.  If the method has no __call__ method,
        then transform is attempted to be called. If this is absent, then fit_transform
        is attempted.


        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            The Phenonaut dataset or Phenonaut object upon which the transform
            should be applied.
        new_feature_names : Optional[Union[list[str], str]], optional
            List of strings containing the names for the new features. Can also
            be just a single string, which then has numerical suffixes attached
            enumerating the number of new features generated. If None, then
            features are attempted to be named through interrogation of the
            method_name argument. By default None.
        method_name : Optional[str], optional
            When setting features, the history message should describe what was
            done. If None, then the method object/method is interrogated in an
            attempt to automatically deduce the name. By default None.
        """
        if not self._is_callable:
            if self._has_transform:
                self.transform(dataset, new_feature_names=new_feature_names, method_name=method_name)
            elif self._has_fit_transform:
                    self.fit_transform(dataset, new_feature_names=new_feature_names, method_name=method_name)
            else:
                raise ValueError("Supplied object/method is not callable and had no fit_transform or transform method")
        if not isinstance(dataset, (Phenonaut, Dataset)):
            raise TypeError(
                f"Dataset should be a Phenonaut object, or phenonaut.Dataset object, it was {type(dataset)}"
            )
        if isinstance(dataset, Phenonaut):
            dataset = dataset.ds
        if self._callable_kwargs == {} or self._callable_kwargs is None:
            transformed_data_array = np.array(self._method(dataset.data))
        else:
            transformed_data_array = np.array(self._method(dataset.data, **self._callable_kwargs))

        method_name, new_feature_names = self._get_method_name_and_feature_names(
            transformed_data_array.shape[1], method_name, new_feature_names
        )
        transformed_df = pd.DataFrame(
            data=transformed_data_array, columns=new_feature_names, index=dataset.df.index
        )
        dataset.df = pd.concat([dataset.df, transformed_df], axis=1)
        dataset.features = (
            new_feature_names,
            f"Applied callable {method_name}, {self._callable_kwargs=}",
        )



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from pathlib import Path
import pandas as pd
from typing import Union
import numpy as np


class PlatemapQuerier:
    plate_to_cpd_to_well_dict = {}
    platemap_files = None

    def __init__(
        self,
        platemap_directory: Union[str, Path],
        platemap_csv_files: Union[list, str, Path] = None,
        plate_name_before_underscore_in_filename=True,
    ):

        self.platemap_files = []
        if platemap_directory is None:
            if platemap_csv_files is None:
                raise FileNotFoundError("No platemaps given")
            if isinstance(platemap_csv_files, str):
                platemap_csv_files = [Path(platemap_csv_files)]
            if isinstance(platemap_csv_files, Path):
                platemap_csv_files = [platemap_csv_files]
            for i, f in enumerate(platemap_csv_files):
                if isinstance(f, str):
                    platemap_csv_files[i] = Path(f)
            for f in platemap_csv_files:
                if not platemap_csv_files[i].exists():
                    raise FileNotFoundError(f"Could not find specified platemap: {f}")
            self.platemap_files = platemap_csv_files
        else:
            if isinstance(platemap_directory, str):
                platemap_directory = Path(platemap_directory)
            if not platemap_directory.is_dir():
                raise FileNotFoundError(f"{platemap_directory} is not a directory")
            self.platemap_files = list(platemap_directory.glob("*.csv"))

        for csv_file in self.platemap_files:
            plate_name = csv_file.stem
            if plate_name_before_underscore_in_filename:
                plate_name = plate_name.split("_")[0]
            df = pd.read_csv(csv_file, index_col=0).rename(columns={"Unnamed: 0": "Column"})
            self.plate_to_cpd_to_well_dict[plate_name] = self._platemap_to_dict(df)

        if len(self.plate_to_cpd_to_well_dict.keys()) == 0:
            raise FileNotFoundError("No suitable platemaps found")

    def _platemap_to_dict(self, plate):
        d = {}
        for column_label, column in plate.iteritems():
            for row, val in enumerate(column):
                if str(val) == "nan":
                    continue
                if val not in d.keys():
                    d[val] = []
                d[val].append(
                    (plate.index.values[row], plate.columns.values[int(column_label) - 1])
                )
        return d

    def get_compound_locations(self, cpd, plates: Union[str, list] = None):
        locations = []
        if isinstance(plates, str):
            plates = [plates]
        for platename, cpd_to_loc_dict in self.plate_to_cpd_to_well_dict.items():
            if plates is not None:
                if not platename in plates:
                    continue
            for location in cpd_to_loc_dict.get(cpd, []):
                locations.append((platename, location[0], location[1]))
        return locations



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from copy import deepcopy
import shutil
from signal import pause
from time import sleep
import pytest
import phenonaut
import pandas as pd
import numpy as np
from iris import *
from phenonaut.data.dataset import Dataset
from phenonaut.output.heatmap import write_heatmap_from_df
from phenonaut.phenonaut import Phenonaut
from pathlib import Path
from shutil import rmtree
import tempfile
from iris import dataset_iris,phenonaut_object_iris_2_views
def test_get_y_from_iris_dataset(dataset_iris):
    from phenonaut.predict.predict_utils import get_y_from_target
    assert get_y_from_target(dataset_iris).shape == (150,)


    
def test_write_heatmap():
    with tempfile.TemporaryDirectory() as tmp_dir:
        df=pd.DataFrame(data=np.array([[0.1,0.2,0.3],[0.1,0.2,0.3],[0.5,0.6,0.7]]), columns=["X1","X2","X3"], index=["Y1", "Y2", "Y3"])
        write_heatmap_from_df(df,"Test heatmap", Path(tmp_dir)/"tmp_heatmap.png")
        assert (Path(tmp_dir)/"tmp_heatmap.png").stat().st_size > 10000



def test_instantiate_dave():
    from phenonaut.predict.default_predictors.pytorch_models.dave import _DAVE_model
    d=_DAVE_model(view_sizes=(100, 150),embedding_size=32, n_hidden=2) 
    assert d.topology is not None


def test_predict_iris_class_1_view(phenonaut_object_iris_2_views, dataset_iris):
    dirpath = tempfile.mkdtemp()
    print(dirpath)
    phe=Phenonaut(dataset_iris)
    import phenonaut.predict    
    phenonaut.predict.profile(phe, dirpath)
    shutil.rmtree(dirpath)


def test_predict_iris_class_2_views(phenonaut_object_iris_2_views):
    dirpath = tempfile.mkdtemp()
    print(dirpath)
    phe=phenonaut_object_iris_2_views
    import phenonaut.predict    
    phenonaut.predict.profile(phe,dirpath)
    shutil.rmtree(dirpath)


# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from phenonaut.data import Dataset

def show_pair_plot(self, ds:Dataset):
    """Visualise the dataset using a pairplot.

    Parameters
    ----------
    ds : Dataset
        The dataset to visualise
    """        
    import matplotlib.pyplot as plt
    import seaborn as sns
    sns.pairplot(ds.df)
    plt.show()



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

import pytest
import phenonaut
import pandas as pd
import phenonaut.transforms
from pathlib import Path
def test_transforms_standardscaler():
    df=pd.DataFrame({
        'ROW':[1,1,1,1,1,1],
        'COLUMN':[1,1,1,1,2,2],
        'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
        'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
        'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
        'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
        'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
        'FOV':[1,2,1,2,1,2]})
    
    phe=phenonaut.Phenonaut(df)
    stdscaler=phenonaut.transforms.StandardScaler()
    stdscaler(phe.ds)
    assert phe.df.iloc[3,4]-1.589652<0.00001

def test_transforms_pca():
    df=pd.DataFrame({
        'ROW':[1,1,1,1,1,1],
        'COLUMN':[1,1,1,1,2,2],
        'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
        'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
        'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
        'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
        'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
        'FOV':[1,2,1,2,1,2]})
    
    phe=phenonaut.Phenonaut(df)
    t_pca=phenonaut.transforms.PCA()
    t_pca(phe.ds)
    assert phe.ds.data.iloc[3,0]-6.829846<0.00001

def test_transforms_tnse():
    df=pd.DataFrame({
        'ROW':[1,1,1,1,1,1],
        'COLUMN':[1,1,1,1,2,2],
        'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
        'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
        'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
        'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
        'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
        'FOV':[1,2,1,2,1,2]})
    
    phe=phenonaut.Phenonaut(df)
    t_tsne=phenonaut.transforms.TSNE()
    t_tsne(phe.ds)
    # Due to the stochastic nature of t-SNE, would not be a good idea to try to compare expected values
    assert len(phe.ds.features)==2

    
def test_transforms_umap():
    from phenonaut import Phenonaut
    df=pd.DataFrame({
        'ROW':[1,1,1,1,1,1],
        'COLUMN':[1,1,1,1,2,2],
        'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
        'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
        'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
        'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
        'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
        'FOV':[1,2,1,2,1,2]})
    
    phe=Phenonaut(df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    t_umap=phenonaut.transforms.UMAP()
    t_umap(phe.ds)
    assert len(phe.ds.features)==2


def test_transforms_custom_callable_square():
    from numpy import square, all as np_all
    from phenonaut import Phenonaut
    df=pd.DataFrame({
        'ROW':[1,1,1,1,1,1],
        'COLUMN':[1,1,1,1,2,2],
        'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
        'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
        'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
        'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
        'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
        'FOV':[1,2,1,2,1,2]})
    
    phe=Phenonaut(df, "PheSmallDS", metadata={'features_prefix':'feat_'})
    f1_series=phe.ds.data.values[:,0]
    t_square=phenonaut.transforms.Transformer(square)
    t_square(phe.ds)
    assert np_all(f1_series*f1_series==phe.ds.data.values[:,0])

def test_transforms_robustmad():
    df=pd.DataFrame({
        'ROW':[1,1,1,1,1,1],
        'COLUMN':[1,1,1,1,2,2],
        'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
        'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
        'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
        'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
        'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
        'FOV':[1,2,1,2,1,2]})
    
    phe=phenonaut.Phenonaut(df)
    robust_mad=phenonaut.transforms.RobustMAD()
    robust_mad(phe.ds)
    assert phe.df.iloc[3,4]-4.363636<0.00001


# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import List, Optional, Union, Tuple
from pathlib import Path
from pandas.errors import DataError

from phenonaut import data, output
from matplotlib import pyplot as plt
import seaborn as sns
from phenonaut import Phenonaut
from phenonaut.data import Dataset

from .visualisation_base import PhenonautVisualisation
import pandas as pd


class Scatter(PhenonautVisualisation):
    """Phenonaut scatter visualisation object.

        The object should be constructed, and then its add member function used
        to add points to the scatter.  It may then be saved by calling save, or
        displayed using show.

        Scatter inherits from the PhenonautVisualisation base class, which
        optionally stores a dictionary called plot_config. As currently
        implemented, plot_config dictionaries can contain one nested dictionary
        under a 'plot_markers' key, with keys to this nested dictionary
        being  perturbation ids, and values are arguments to the matplotlib
        marker types. To populate this base class dictionary, pass a config
        dictionary to the constructor, like
        
        .. code-block:: python
        
            {'plot_markers':{'pert1':'X', 'pert2':'x'}}.

        Parameters
        ----------
        plot_config : Optional[Union[Path, str, dict]], optional
            Optional configuration dictionary, allows  alows specification of initialisation arguments via dictionary.
            Supply keys and values as argument as values. By default None.
        figsize : Optional[Tuple[float, float]], optional
            Output figure size (in inches, as directed by matplotlib/seaborn).
            By default (8, 6)
        title : Optional[str], optional
            Plot title, by default "2D scatter"
        x_label : Optional[str], optional
            x-axis label, by default None
        y_label : Optional[str], optional
            y-axis label, by default None
        show_legend : Optional[bool], optional
            If True, then add a legend to the plot. By default True.
        marker_size : int, optional
            The size of each datapoint within the scatter. By default 90
        axis_ranges : Optional[Tuple[Tuple[float, float], Tuple[float, float]]], optional
            Optional tuple of tuples giving the min and max extents of each axis.
            If None is given for any value, then the corresponding min/max for
            that axis is calcaulated and used.
            By default ((None, None), (None, None)).
        """
    
    def __init__(
        self,
        plot_config: Optional[Union[Path, str, dict]] = None,
        figsize: Optional[Tuple[float, float]] = (8, 6),
        title: Optional[str] = "2D scatter",
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
        show_legend: Optional[bool] = True,
        marker_size: int = 90,
        axis_ranges: Optional[Tuple[Tuple[float, float], Tuple[float, float]]] = ((None, None), (None, None)),
    ):
        """Phenonaut scatter visualisation object.

        The object should be constructed, and then its add member function used
        to add points to the scatter.  It may then be saved by calling save, or
        displayed using show.

        Scatter inherits from the PhenonautVisualisation base class, which
        optionally stores a dictionary called plot_config. As currently
        implemented, plot_config dictionaries can contain one nested dictionary
        under a 'plot_markers' key, with keys to this nested dictionary
        being  perturbation ids, and values are arguments to the matplotlib
        marker types. To populate this base class dictionary, pass a config
        dictionary to the constructor, like
        
        .. code-block:: python
        
            {'plot_markers':{'pert1':'X', 'pert2':'x'}}.

        Parameters
        ----------
        plot_config : Optional[Union[Path, str, dict]], optional
            Optional configuration dictionary, allows  alows specification of initialisation arguments via dictionary.
            Supply keys and values as argument as values. By default None.
        figsize : Optional[Tuple[float, float]], optional
            Output figure size (in inches, as directed by matplotlib/seaborn).
            By default (8, 6)
        title : Optional[str], optional
            Plot title, by default "2D scatter"
        x_label : Optional[str], optional
            x-axis label, by default None
        y_label : Optional[str], optional
            y-axis label, by default None
        show_legend : Optional[bool], optional
            If True, then add a legend to the plot. By default True.
        marker_size : int, optional
            The size of each datapoint within the scatter. By default 90
        axis_ranges : Optional[Tuple[Tuple[float, float], Tuple[float, float]]], optional
            Optional tuple of tuples giving the min and max extents of each axis.
            If None is given for any value, then the corresponding min/max for
            that axis is calcaulated and used.
            By default ((None, None), (None, None)).
        """
        super().__init__(plot_config)
        # Add arguments to figure config
        for var in [
            x
            for x in self.__init__.__code__.co_varnames
            if x not in ("self", "figure_config", "var") and x not in self.config.keys()
        ]:
            self.config[var] = eval(var)
        self.fig, self.ax = plt.subplots(1, 1, figsize=figsize)
        self.marker_size = marker_size

    def add(
        self,
        dataset: Union[Dataset, Phenonaut],
        perturbations: Optional[List[str]] = None,
        markers: Union[dict, list, bool] = True,
        marker_size:int=None,
    ):
        """Add points to the scatter plot

        Parameters
        ----------
        dataset : Union[Dataset, Phenonaut]
            Phenonaut dataset or Phenonaut object (if it contains only one
            dataset) from which to take datapoints.
        perturbations : Optional[List[str]], optional
            If perturbation column is set on the dataset and a list of
            perturbations is given here, then plot only those matching
            datapoints.  If None, then all datapoints are plotted.
            By default None,
        markers : Union[dict, list, bool], optional
            Dictionary containing pertubation ids as keys, and then values to
            pass to matplotlib defining marker styles, or a list of tuples of
            this form. If True, then let matplotlib enumerate marker types.
            By default True.
        marker_size : int, optional
            Size at which to draw datapoints. If None, then the value is taken
            from the marker_size argument used to construct the scatter object
            see constructor docstring. By default None.

        Raises
        ------
        DataError
            Scatter plots are 2D, more than 2 features found.
        """    
        if isinstance(dataset, Phenonaut):
            dataset=dataset[-1]
        if len(dataset.features) != 2:
            raise DataError(f"Scatter requires 2 features, got {len(dataset.features)}")
        if self.config.get("xlabel") is None:
            self.config["xlabel"] = dataset.features[0]
        if self.config.get("ylabel") is None:
            self.config["ylabel"] = dataset.features[1]

        if marker_size is None:
            marker_size=self.marker_size

        if perturbations is None:
            perturbations = dataset.get_unique_perturbations()
        if perturbations is not None:
            df = dataset.df.query(f"{dataset.perturbation_column} == @perturbations")
        else:
            df = dataset.df
        sns.scatterplot(
            data=df,
            x=dataset.features[0],
            y=dataset.features[1],
            style=dataset.perturbation_column,
            hue=dataset.perturbation_column,
            s=marker_size,
            markers=markers,
            ax=self.ax,
        )

    def _decorate_figure(
        self,
    ):
        """Internal helper function for axis decoration
        """        
        if self.config.get("show_legend", True):
            self.ax.legend()
        else:
            self.ax.get_legend().remove()

        self.ax.set_xlabel(self.config.get("xlabel", ""))
        self.ax.set_ylabel(self.config.get("ylabel", ""))
        axis_ranges = self.config.get("axis_ranges", ((None, None), (None, None)))
        self.ax.set_xlim(axis_ranges[0])
        self.ax.set_ylim(axis_ranges[1])
        self.ax.set_title(self.config.get("title", "2D scatter"))
        plt.tight_layout()

    def show(
        self,
    ):
        """Show the plot on the screen
        """        
        self._decorate_figure()
        plt.show()

    def save_figure(self, output_image_path: Union[Path, str], **savefig_kwargs):
        """Save the current scatter plot to PNG/SVG file

        Saving of the image is achieved using the matplotlib plt.savefig. After
        the output filename, any argument/option may be given that is also
        valid in calls to plt.savefig.

        Parameters
        ----------
        output_image_path : Union[Path, str]
            Output file path for the scatter plot.
        """        
        self._decorate_figure()
        if isinstance(output_image_path, str):
            output_image_path = Path(output_image_path)
        if not output_image_path.parent.exists():
            output_image_path.parent.mkdir(parents=True)
        plt.savefig(output_image_path, **savefig_kwargs)
    
    def __del__(self,):
        """Delete scatter

        Ensures plt.cla and clf are called upon Scatter object deletion
        """        
        plt.cla()
        plt.clf()


# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from time import sleep
import pytest
import phenonaut
from sklearn.datasets import load_iris
from iris import *
from phenonaut.data.dataset import Dataset
import pandas as pd
import numpy as np
from phenonaut.phenonaut import Phenonaut
from io import StringIO

def test_get_features_and_non_features(dataset_iris):
    features = dataset_iris.features
    print(features)
    non_features = dataset_iris.get_non_feature_columns()
    assert sorted(features + non_features) == sorted(dataset_iris.df.columns.values)

def test_new_aggregated_dataset():
    """Here we test a df as follows:
    ROW	COLUMN	BARCODE	feat_1	feat_2	feat_3	filename	FOV
    1	1	    Plate1	1.2	    1.2	    1.3	    fileA.png	1
    1	1	    Plate1	1.3	    1.4	    1.5	    FileB.png	2
    1	1	    Plate2	5.2	    5.1	    5	    FileC.png	1
    1	1	    Plate2	6.2	    6.1	    6.8	    FileD.png	2
    1	2	    Plate1	0.1	    0.2	    0.3	    fileE.png	1
    1	2	    Plate1	0.2	    0.2	    0.38    FileF.png	2
    
    As there are multiple fields of view, we merge on common ROW, COLUMN and BARCODE fields.
    Merging uses np.mean, but may be any callable which returns a sing value 
    
    """
    
    df=pd.DataFrame({
        'ROW':[1,1,1,1,1,1],
        'COLUMN':[1,1,1,1,2,2],
        'BARCODE':["Plate1","Plate1","Plate2","Plate2","Plate1","Plate1"],
        'feat_1':[1.2,1.3,5.2,6.2,0.1,0.2],
        'feat_2':[1.2,1.4,5.1,6.1,0.2,0.2],
        'feat_3':[1.3,1.5,5,6.8,0.3,0.38],
        'filename':['fileA.png','FileB.png','FileC.png','FileD.png','fileE.png','FileF.png'],
        'FOV':[1,2,1,2,1,2]})
    
    phe=phenonaut.Phenonaut(df, metadata={'features_prefix':'feat_'})
    new_ds=phe.ds.new_aggregated_dataset(["ROW", "COLUMN", "BARCODE"])
    print(new_ds.df)
    assert set(new_ds.features)==set(['feat_1','feat_2','feat_3'])
    assert abs(new_ds.df['feat_1'][0]-1.25<1e-6)
    assert np.abs(new_ds.df['feat_1'][1]-5.70<1e-6)
    assert np.abs(new_ds.df['feat_1'][2]-0.15<1e-6)
    assert np.abs(new_ds.df['feat_2'][0]-1.3<1e-6)
    assert np.abs(new_ds.df['feat_2'][1]-5.6<1e-6)
    assert np.abs(new_ds.df['feat_2'][2]-0.2<1e-6)
    assert np.abs(new_ds.df['feat_3'][0]-1.4<1e-6)
    assert np.abs(new_ds.df['feat_3'][1]-5.9<1e-6)
    assert np.abs(new_ds.df['feat_3'][2]-0.34<1e-6)
    

def test_iris_packageddataset():
    import tempfile

    tmpdir=tempfile.gettempdir()
    from phenonaut.packaged_datasets import Iris, Iris_2_views
    phe=Phenonaut(Iris(tmpdir))
    assert len(phe.keys())==1


def test_feature_selection_when_loading_csv():
    import tempfile
    with tempfile.NamedTemporaryFile(mode = "w") as tmp:
        tmp.write("row, column, feature1, feature2, derived_feature3, feature10, feature003\n1,1,1,2,3,4,5\n2,1,6,7,8,9,10")
        tmp.flush()
        phe=Phenonaut()
        phe.load_dataset(tmp.name,tmp.name, {'features_prefix':"feat"})
        assert phe.ds.features==['feature1', 'feature2', 'feature003', 'feature10']
        
    with tempfile.NamedTemporaryFile(mode = "w") as tmp:
        tmp.write("sepal length (cm),sepal width (cm),petal length (cm),petal width (cm),target\n5.4,3.4,1.7,0.2,0\n7.2,3.0,5.8,1.6,2\n6.4,2.8,5.6,2.1,2\n4.8,3.1,1.60.2,0\n5.6,2.5,3.9,1.1,1")
        tmp.flush()
        phe=Phenonaut()
        phe.load_dataset(tmp.name,tmp.name, {'features_regex':".*(width|length).*"})
        assert phe.ds.features==['petal length (cm)', 'petal width (cm)', 'sepal length (cm)', 'sepal width (cm)']


def test_replace_string_in_column():
    """Here we test a df as follows:
    ROW	COLUMN	BARCODE	feat_1	feat_2	feat_3	filename	FOV
    1	1	    Plate1	1.2	    1.2	    1.3	    fileA.png	1
    1	1	    Plate1	1.3	    1.4	    1.5	    FileB.png	2
    1	1	    Plate2	5.2	    5.1	    5	    FileC.png	1
    1	1	    Plate2	6.2	    6.1	    6.8	    FileD.png	2
    1	2	    Plate1	0.1	    0.2	    0.3	    fileE.png	1
    1	2	    Plate1	0.2	    0.2	    0.38    FileF.png	2
    
    """

    df=pd.read_csv(StringIO(
"""ROW,COLUMN,BARCODE,feat_1,feat_2,feat_3,filename,FOV
1,1,Plate1,1.2,1.21.3,fileA.png,1
1,1,Plate1,1.3,1.41.5,FileB.png,2
1,1,Plate2,5.2,5.15,FileC.png,1
1,1,Plate2,6.2,6.16.8,FileD.png,2
1,2,Plate1,0.1,0.20.3,fileE.png,1
1,2,Plate1,0.2,0.20.38,FileF.png,2
"""))
    
    phe=phenonaut.Phenonaut(df, metadata={'features_prefix':'feat_'})
    phe[-1].replace_str("BARCODE", "Plate", "P")
    assert all(elem in ['P1', 'P2'] for elem in phe.df.BARCODE.unique())




# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from typing import Any, Iterable, Optional, Type, Union, Callable
from dataclasses import dataclass, field
from sklearn.base import BaseEstimator
from sklearn.metrics import accuracy_score
from sklearn.metrics import mean_absolute_error
from functools import reduce

@dataclass
class OptunaHyperparameter:
    """OptunaHyperparameter base dataclass which is inherited from
    """    
    name: str

@dataclass
class OptunaHyperparameterNumber(OptunaHyperparameter):
    """Optuna hyperparameter dataclass for numbers, inherited for int and float


    Raises
    ------
    ValueError
        Lower bound must be lower than upper bound.
    """    
    lower_bound: int
    upper_bound: int
    needed:bool=True
    def _check_bounds(self):
        if self.upper_bound <= self.lower_bound:
            raise ValueError(
                f"lower_bound must be lower than upper_bound "
                f"({self.lower_bound}!<{self.upper_bound}), consider a"
                "categorical hyperparameter with one option if you need"
                " a single choice optuna hyperparameter"
            )

@dataclass
class HyperparameterInt(OptunaHyperparameterNumber):
    """Optuna hyperparameter dataclass for ints

    """    
    def __post_init__(self):
        self._check_bounds()
        self.optuna_func = "suggest_int"
        self.parameters = (self.lower_bound, self.upper_bound)


@dataclass
class HyperparameterFloat(OptunaHyperparameterNumber):
    """Optuna hyperparameter dataclass for floats

    """    
    def __post_init__(self):
        self._check_bounds()
        self.optuna_func = "suggest_float"
        self.parameters = (self.lower_bound, self.upper_bound)

@dataclass
class HyperparameterLog(OptunaHyperparameterNumber):
    """Optuna hyperparameter dataclass for loguniform distributions of floats

    """ 
    def __post_init__(self):
        self._check_bounds()
        self.optuna_func = "suggest_loguniform"
        self.parameters = (self.lower_bound, self.upper_bound)


@dataclass
class HyperparameterCategorical(OptunaHyperparameter):
    """Optuna hyperparameter dataclass for categorical lists

    """ 
    choices: Union[list, tuple]
    needed:bool=True
    def __post_init__(self):
        if len(self.choices)==0:
            raise ValueError("Choices must be an iterable (usualy list or tuple) with length at least one")
        self.optuna_func = "suggest_categorical"
        self.parameters = (self.choices,)


@dataclass
class PhenonautPredictionMetric:
    """PhenonautPredictionMetric dataclass to hold metric, name and direction.

    """ 
    func: Callable
    name: str
    lower_is_better: bool
    def __call__(self, *args: Any, **kwds: Any) -> float:
        return self.func(*args, **kwds)

@dataclass(unsafe_hash=True)
class PhenonautPredictor:
    """PhenonautPredictor dataclass

    The PhenonautPredictor wraps classes with fit and predict methods,
    augmenting them with additional information like name, the number of views
    it may operate on at once, and hyperparameter lists which may be optimised
    using Optuna.
    """    
    name: str
    predictor: Union[BaseEstimator, Callable]
    optuna: Optional[Union[Iterable[OptunaHyperparameter], OptunaHyperparameter]] = None
    num_views: int = 1
    max_optuna_trials: Optional[int] = None
    dataset_size_cutoff: Optional[int] = None
    constructor_kwargs:dict=field(default_factory=dict)
    max_classes: Optional[int] = None
    conditional_hyperparameter_generator_constructor_keyword:Optional[str]=None,
    conditional_hyperparameter_generator:Optional[Callable]=None
    embed_in_results:bool = True

    # Standardise self.optuna to an iterable if required.
    def __post_init__(self):
        if isinstance(self.optuna, OptunaHyperparameter):
            self.optuna = (self.optuna,)
        if self.optuna is None and self.max_optuna_trials is None:
            self.max_optuna_trials = 1
        if isinstance(self.optuna, Iterable):
            if all([isinstance(opt_option, HyperparameterCategorical) for opt_option in self.optuna]):
                self.max_optuna_trials = reduce(
                    lambda x, y: x * y, [len(opt_options.parameters) for opt_options in self.optuna]
                )



# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from collections import namedtuple
from pathlib import Path
import pathlib
from typing import List, NamedTuple, Optional, Union
from ..data import Dataset
import pandas as pd
from .base import PackagedDataset
import gzip, tarfile
import shutil
from sklearn.datasets import load_iris

import pandas as pd
import numpy as np
import h5py
import datetime


class Iris(PackagedDataset):
    """IRIS dataset Scikit learn

        This PackagedDataset provides the Iris dataset from scikit-learn, with
        unique iris entries, each with four features each, and finally a target
        column denoting class. Further information is available via scikit-lean
        here:

        https://scikit-learn.org/stable/auto_examples/datasets/plot_iris_dataset.html

        or wikipedia:

        https://en.wikipedia.org/wiki/Iris_flower_data_set

        or from:

        https://archive.ics.uci.edu/ml/datasets/Iris

        Parameters
        ----------
        root : Union[Path, str]
            Local directory containing the prepared dataset. If the dataset is
            not found here and the argument download=True is not given, then an
            error is raised. If download=True and the processed dataset is
            absent, then it is downloaded the directory pointed at by the
            'raw_data' argument detailed below. If raw_data_dir is a
            non-absolute path, such as a single directory, then it is created as
            a subdirectory of this root directory.
        download : bool, optional
            If true and the processed dataset is not found in the root
            directory, then the dataset is downloaded and processed.
            By default False.
        raw_data_dir : Optional[Union[Path, str]], optional
            If downloading and preparing the dataset, then a directory for the
            raw data may be specified. If a non-absolute location is given, then
            it is created in a subdirectory of the root directory specified as
            the first argument. Absolute paths may be used to place raw
            datafiles and intermediates in another location, such as scratch
            disks etc, by default Path("raw_data").
        """
    def __init__(
        self,
        root: Union[Path, str],
        download: bool = False,
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        rm_downloaded_data: bool = True,
    ):

        super().__init__(root, raw_data_dir)
        self.name = "Iris"
        self.processed_h5_file = self.root / "IRIS_phenonaut.h5"

        # If the dataset is missing, get it
        self._call_if_file_missing(self.processed_h5_file, self._make, None)

        self.store = pd.HDFStore(self.processed_h5_file)

        self.register_ds_key("Iris")

        self.store.close()

    def _make(self, remove_intermediates=False):
        """Make IRIS HDF5 file

        Parameters
        ----------
        remove_intermediates : bool, optional
            If true, then the downloaded archive is removed, by default False
        """
        iris_df = load_iris(as_frame=True).frame
        h5_store = pd.HDFStore(self.processed_h5_file, "w", complevel=9)

        h5_store["/Iris"] = iris_df

        h5_store["/creation_date"] = pd.Series(str(datetime.datetime.now()))

        h5_store.close()

    def get_df(self, key: str) -> pd.DataFrame:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of pd.DataFrame.

        Returns
        -------
        pd.DataFrame
            Requested pd.DataFrame from h5 store
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        store.close()
        return df

    def get_ds(self, key: str) -> Dataset:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of Phenonaut Dataset.

        Returns
        -------
        Dataset
            Requested Phenonaut Dataset from h5 store, with the correctly set
            features and metadata
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        features = [f for f in list(df.columns) if f not in ["target"]]
        ds = Dataset("Iris", df, {"features": features})
        store.close()
        return ds


class Iris_2_views(PackagedDataset):
    def __init__(
        self,
        root: Union[Path, str],
        download: bool = False,
        raw_data_dir: Optional[Union[Path, str]] = Path("raw_data"),
        rm_downloaded_data: bool = True,
    ):
        """IRIS dataset Scikit learn - 2 views

        This PackagedDataset provides 2 views of the Iris dataset from
        scikit-learn, with unique iris entries, each with four features each,
        and finally a target column denoting class. Critically, this
        PackagedDataset loader contains two views, splitting the four feature
        columns between views.

        Further
        information is available via scikit-lean here:
        https://scikit-learn.org/stable/auto_examples/datasets/plot_iris_dataset.html

        or wikipedia:
        https://en.wikipedia.org/wiki/Iris_flower_data_set

        or from
        https://archive.ics.uci.edu/ml/datasets/Iris

        Parameters
        ----------
        root : Union[Path, str]
            Local directory containing the prepared dataset. If the dataset is
            not found here and the argument download=True is not given, then an
            error is raised. If download=True and the processed dataset is
            absent, then it is downloaded the directory pointed at by the
            'raw_data' argument detailed below. If raw_data_dir is a
            non-absolute path, such as a single directory, then it is created
            as a subdirectory of this root directory.
        download : bool, optional
            If true and the processed dataset is not found in the root
            directory, then the dataset is downloaded and processed.
            By default False.
        raw_data_dir : Optional[Union[Path, str]], optional
            If downloading and preparing the dataset, then a directory for the
            raw data may be specified. If a non-absolute location is given, then
            it is created in a subdirectory of the root directory specified as
            the first argument. Absolute paths may be used to place raw
            datafiles and intermediates in another location, such as scratch
            disks etc, by default Path("raw_data").
        """

        super().__init__(root, raw_data_dir)
        self.name = "Iris"
        self.processed_h5_file = self.root / "IRIS_2_views_phenonaut.h5"

        # If the dataset is missing, get it
        self._call_if_file_missing(self.processed_h5_file, self._make, None)

        self.store = pd.HDFStore(self.processed_h5_file)

        self.register_ds_key("Iris_view_1")
        self.register_ds_key("Iris_view_2")

        self.store.close()

    def _make(self, remove_intermediates=False):
        """Make IRIS HDF5 file

        Parameters
        ----------
        remove_intermediates : bool, optional
            If true, then the downloaded archive is removed, by default False
        """
        iris_df = load_iris(as_frame=True).frame
        h5_store = pd.HDFStore(self.processed_h5_file, "w", complevel=9)

        df1 = iris_df.iloc[:, [0, 1, 4]].copy()
        df2 = iris_df.iloc[:, [2, 3, 4]].copy()

        h5_store["/Iris_view_1"] = df1
        h5_store["/Iris_view_2"] = df2
        h5_store["/creation_date"] = pd.Series(str(datetime.datetime.now()))
        self.register_ds_key("Iris_view_1")
        h5_store.close()

    def get_df(self, key: str) -> pd.DataFrame:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of pd.DataFrame.

        Returns
        -------
        pd.DataFrame
            Requested pd.DataFrame from h5 store
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        store.close()
        return df

    def get_ds(self, key: str) -> Dataset:
        """Get supporting dataframe

        Parameters
        ----------
        key : str
            Key of Phenonaut Dataset.

        Returns
        -------
        Dataset
            Requested Phenonaut Dataset from h5 store, with the correctly set
            features and metadata
        """
        store = pd.HDFStore(self.processed_h5_file)
        df = store["/" + key]
        features = [f for f in list(df.columns) if f not in ["target"]]
        ds = Dataset(key, df, {"features": features})
        store.close()
        return ds
